from collections import defaultdict
from collections.abc import Sequence, MutableMapping, Mapping
import functools
import weakref
import re
import numbers

import numpy as np

from brian2.core.base import weakproxy_with_fallback
from brian2.core.base import device_override
from brian2.core.namespace import get_local_namespace
from brian2.core.variables import (DynamicArrayVariable, Variables)
from brian2.codegen.codeobject import create_runner_codeobj
from brian2.codegen.translation import get_identifiers_recursively
from brian2.devices.device import get_device
from brian2.equations.equations import (Equations,
                                        DIFFERENTIAL_EQUATION, SUBEXPRESSION,
                                        PARAMETER,
                                        check_subexpressions, EquationError)
from brian2.groups.group import Group, CodeRunner, get_dtype
from brian2.groups.neurongroup import (extract_constant_subexpressions,
                                       SubexpressionUpdater,
                                       check_identifier_pre_post)
from brian2.parsing.expressions import is_boolean_expression, parse_expression_dimensions
from brian2.stateupdaters.base import (StateUpdateMethod,
                                       UnsupportedEquationsException)
from brian2.stateupdaters.exact import linear, independent
from brian2.units.fundamentalunits import (Quantity, DIMENSIONLESS, DimensionMismatchError,
                                           fail_for_dimension_mismatch)
from brian2.units.allunits import second
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import get_identifiers, word_substitute
from brian2.utils.arrays import calc_repeats
from brian2.core.spikesource import SpikeSource
from brian2.synapses.parse_synaptic_generator_syntax import parse_synapse_generator
from brian2.parsing.bast import brian_ast
from brian2.parsing.rendering import NodeRenderer
MAX_SYNAPSES = 2147483647


def spike_queue(self, source_start, source_end):
    """
    TODO

    Parameters
    ----------
    source_start
        TODO
    source_end
        TODO
    
    Returns
    -------
    `SpikeQueue`
        TODO
    """

    # TODO REMOVE! In brian2lava there is no need for a spike queue.
    # Use the C++ version of the SpikeQueue when available
    try:
        from brian2.synapses.cythonspikequeue import SpikeQueue
        self.logger.diagnostic('Using the C++ SpikeQueue', once=True)
    except ImportError:
        from brian2.synapses.spikequeue import SpikeQueue
        self.logger.diagnostic('Using the Python SpikeQueue', once=True)

    return SpikeQueue(source_start=source_start, source_end=source_end)


def synapses_connect(
        self, synapses, condition=None, i=None, j=None, p=1., n=1,
        skip_if_invalid=False, namespace=None, level=0
    ):
    """
    Connects synapses.

    This method overwrites the `connect` function from Brian.

    Parameters
    ----------
    synapses : `Synapses`
        Equals the `self` of the original `connect` function from Brian,
        which is an instance of the `Synapses` class.
    condition : str, bool, optional
        A boolean or string expression that evaluates to a boolean.
        The expression can depend on indices ``i`` and ``j`` and on
        pre- and post-synaptic variables. Can be combined with
        arguments ``n``, and ``p`` but not ``i`` or ``j``.
    i : int, ndarray of int, str, optional
        The presynaptic neuron indices  It can be an index or array of
        indices if combined with the ``j`` argument, or it can be a string
        generator expression.
    j : int, ndarray of int, str, optional
        The postsynaptic neuron indices. It can be an index or array of
        indices if combined with the ``i`` argument, or it can be a string
        generator expression.
    p : float, str, optional
        The probability to create ``n`` synapses wherever the ``condition``
        evaluates to true. Cannot be used with generator syntax for ``j``.
    n : int, str, optional
        The number of synapses to create per pre/post connection pair.
        Defaults to 1.
    skip_if_invalid : bool, optional
        If set to True, rather than raising an error if you try to
        create an invalid/out of range pair (i, j) it will just
        quietly skip those synapses.
    namespace : dict-like, optional
        A namespace that will be used in addition to the group-specific
        namespaces (if defined). If not specified, the locals
        and globals around the run function will be used.
    level : int, optional
        How deep to go up the stack frame to look for the locals/global
        (see ``namespace`` argument).
    """

    # Check types
    synapses._verify_connect_argument_types(condition, i, j, n, p)

    synapses._connect_called = True

    # Get namespace information
    if namespace is None:
        namespace = get_local_namespace(level=level + 2)

    try:  # wrap everything to catch IndexError
        # which connection case are we in?
        # 1: Connection condition
        if condition is None and i is None and j is None:
            condition = True
        if condition is not None:
            if i is not None or j is not None:
                raise ValueError("Cannot combine condition with i or j "
                                    "arguments")
            if condition is False or condition == 'False':
                # Nothing to do
                return
            j = synapses._condition_to_generator_expression(condition, p, namespace)
            print("Using synapses generator")
            synapses._add_synapses_generator(j, n, skip_if_invalid=skip_if_invalid,
                                            namespace=namespace, level=level + 2,
                                            over_presynaptic=True)
        # 2: connection indices
        elif (i is not None and j is not None) and not (isinstance(i, str) or isinstance(j, str)):
            if skip_if_invalid:
                raise ValueError("Can only use skip_if_invalid with string "
                                    "syntax")
            i, j, n = synapses._verify_connect_array_arguments(i, j, n)
            synapses._add_synapses_from_arrays(i, j, n, p, namespace=namespace)
        # 3: Generator expression over post-synaptic cells (i='...')
        elif isinstance(i, str):
            i = synapses._finalize_generator_expression(i, j, p, 'i', 'j')
            synapses._add_synapses_generator(i, n, skip_if_invalid=skip_if_invalid,
                                            namespace=namespace, level=level + 2,
                                            over_presynaptic=False)
        # 4: Generator expression over pre-synaptic cells (i='...')
        elif isinstance(j, str):
            j = synapses._finalize_generator_expression(j, i, p, 'j', 'i')
            synapses._add_synapses_generator(j, n, skip_if_invalid=skip_if_invalid,
                                            namespace=namespace, level=level + 2,
                                            over_presynaptic=True)
        else:
            raise ValueError("Must specify at least one of condition, i or "
                                "j arguments")
    except IndexError as e:
        raise IndexError("Tried to create synapse indices outside valid "
                            "range. Original error message: " + str(e))
    # Do some refactoring of the synapses code to adapt it to Lava API:
    # FIXME: Probably this will be useless, but I'll wait before deleting it since it does no harm here for now.
    self.add_lava_synapses_generator(synapses)


def add_lava_synapses_generator(self, synapses):
    """
    A helper function to correctly tell lava which synaptic variables will be
    initialized through this generator codeobject.

    Parameters
    ----------
    synapses : `Synapses`
        An instance of the `Synapses` class

    Notes
    -----
    At the moment I'm using a trick, which just adds a random int, used in build() to determine that this codeobject
    initializes multiple variables. In the future this should be changed: either expanded if the variables themselves
    are needed, or replaced by something more elegant.
    """
    # Extract the generator codeobj from the device's code_objects
    # TODO: This is not clean and should be handled in a different way.
    generator_codeobj = [self.code_objects[name] for name in self.code_objects if ('synapses_create' in name and self.code_objects[name].owner == synapses)][-1]
    self.init_variables[generator_codeobj.name] = 123



from brian2.synapses.synapses import SynapticPathway
from brian2.core.functions import Function
from brian2.core.variables import ArrayVariable


def code_object(
        self,
        owner,
        name,
        abstract_code,
        variables,
        template_name,
        variable_indices,
        codeobj_class=None,
        template_kwds=None,
        override_conditional_write=None,
        compiler_kwds=None
    ):
    """
    Defines a code object.

    Parameters
    ----------
    TODO

    Returns
    -------
    codeobj
    """
    
    # Log when a code object is added
    self.logger.diagnostic(f'Add code_object {name}')
    self.logger.diagnostic(f'Variables gotten from previous steps, {[varname for varname,var in list(variables.items())]}')

    # Init template keywords if none were given
    if template_kwds is None:
        template_kwds = dict()

    # FIXME: This will need to be fixed because it can cause problems when we generate multiple objects of the same kind
    template_kwds['name'] = name[:-1] if name[-1] == '*' else name

    # In case a variable is set with initial values, we extract the related variable name
    # The variable name is extracted from the abstract code line (before the equal sign)
    # The name is used to get a unique name for the method that initializes the variable
    if template_name in self.init_template_functions:

        # Find variable to initialize
        lava_variable = None
        for v in variables.values():
            if hasattr(v, 'owner'):
                lava_var_name = self.get_array_name(v, prefix='')
                if lava_var_name in self.init_variables.values():
                    lava_variable = v
                    break

        # TODO raise exception if lava_variable was not found (is still None)

        lava_array_name = self.get_array_name(v, prefix='self.init')
        # For now let's separate the synapses case from the variableview one.
        # TODO: It should be possible to unify the procedure for both, so we should do that..
        if not 'synapses' in name:
            # Add statement to template keywords
            template_kwds['return_statement'] = f'return np.array({lava_array_name})'

            # We take the method that was lastly added from the variableview
            # Francesco: here we can't use simply 'name' because this conflicts with the 
            # arguments to the code_object() function
            method_name = self.last_set_variable_method_name
            # Add method name to template keywords
            template_kwds['name'] = method_name

        # With this we bypass the instructions in the brian base device:
        # (brian2.devices.device -> 328-336)
        # This is because for initialization variables we want to use a different
        # naming convention.
        # Note that this requires renaming the {{variables}} in the template to add the '_init' suffix
        for varname, var in variables.items():
            if isinstance(var, ArrayVariable):
                pointer_name = self.get_array_name(var, prefix = 'self.init')
                if var.scalar:
                    pointer_name += "[0]"
                template_kwds[varname + '_init'] = pointer_name
                if hasattr(var, "resize"):
                    dyn_array_name = self.get_array_name(var, prefix = 'self.init', access_data=False)
                    template_kwds[f"_dynamic_{varname}_init"] = dyn_array_name
        # TODO: This functionality could be used in the other generators (for variable initializations)

    # I think this has to be managed here, because the CodeRunner for synaptic Pathways gets 
    # created during synapses initialization, and thus is out of our control until this point.
    if template_name == 'synapses':
        pathway = template_kwds['pathway']
        self.determine_lava_ports(pathway, {v:var for v,var in variables.items() if not isinstance(var,Function)})
        # Here we only change the name, to make it easier for debugging
        template_kwds_neuron = template_kwds.copy()
        codeobj_name = pathway.target.name +'_'+ owner.name + '_' + pathway.prepost + '_' + 'activation_processing'
        template_kwds_neuron['name'] = codeobj_name
        # Here we substitute the codeobject with one which belongs to the correct NeuronGroup
        codeobj = self.super.code_object(
            pathway.target,
            codeobj_name,
            abstract_code,
            variables,
            'activation_processing',
            variable_indices,
            codeobj_class=codeobj_class,
            template_kwds=template_kwds_neuron,
            override_conditional_write=override_conditional_write,
            compiler_kwds=compiler_kwds
        )
        self.code_objects[codeobj.name] = codeobj
        # Make a different abstract code for synapses
        # FIXME: this is not very clean and should probably be handled differently
        code = ''
        abstract_code = {None : code}
    # Call code_object method from Brian2 parent device
    codeobj = self.super.code_object(
        owner,
        name,
        abstract_code,
        variables,
        template_name,
        variable_indices,
        codeobj_class=codeobj_class,
        template_kwds=template_kwds,
        override_conditional_write=override_conditional_write,
        compiler_kwds=compiler_kwds
    )

    # Store code objects in device
    self.code_objects[codeobj.name] = codeobj
    
    return codeobj


def determine_lava_ports(self, pathway, variables):
    """
    Extracts the synaptic variables to be sent to the neurongroups and determine the Lava ports required for that.

    Parameters
    ----------
    pathway
        Synaptic pathways.
    variables
        All variables that are not a function.
    """

    synaptic_vars = []
    # TODO: this is probably needed in case the update only requires synaptic variables,
    # then you wouldn't need to send them to the neuronGroup. But for now let's just make
    # the easiest scenario
    for varname, var in variables.items():
        if not 'synapses' in var.owner.name:
            continue
        synaptic_vars.append(varname)
    # At the first iteration of this function we create the corresponding dictionary
    if not hasattr(self, 'lava_ports'):
        self.lava_ports = {}
    # Var names are kept in their original form, but the ports are separated depending on the pathway
    for varname in synaptic_vars:
        portname = pathway.synapses.name +'_'+ pathway.prepost +'_'+ varname
        self.lava_ports[portname] = {
            'varname': varname,
            'portname': portname,
            'pathway': pathway,
            'sender' : pathway.source.name,
            'receiver': pathway.target.name,
            'shape': None
        }
    
    







import os
import tempfile
import regex
import numpy as np
from jinja2 import FileSystemLoader, Environment

# Import Brian2 modules
# from brian2.units.allunits import second
# from brian2.core.functions import Function
# from brian2.core.variables import Constant, ArrayVariable, AuxiliaryVariable
from brian2.groups.neurongroup import NeuronGroup, StateUpdater, Resetter, Thresholder, SubexpressionUpdater
from brian2.input.poissongroup import PoissonGroup
from brian2.input.spikegeneratorgroup import SpikeGeneratorGroup
from brian2.monitors.spikemonitor import SpikeMonitor
from brian2.monitors.ratemonitor import PopulationRateMonitor
from brian2.monitors.statemonitor import StateMonitor
from brian2.synapses.synapses import Synapses
# from brian2.units import second

from pprint import pprint


def build(
    self,
    hardware='CPU',
    debug=False,
    direct_call=True
):
    """
    Builds the Lava executables.

    It contains the following steps:

    *   includes some checks
    *   initializes file write
    *   renders the templates
    
    Parameters
    ----------
    hardware : `str`, optional
        The underlying hardware where the code is executed. Defaults to `CPU`.
    debug : `bool`, optional
        Whether to compile in debug mode. Defaults to ``False``.
    direct_call : `bool`, optional
        Whether this function was called directly. Is used internally to
        distinguish an automatic build due to the ``build_on_run`` option
        from a manual ``device.build`` call.

    Notes
    -----
    TODO change default hardware to `Loihi` as soon as Loihi is supported
    """
    
    # Log that the build method was called
    self.logger.debug("Building Lava device.")
    
    # Check if direct_call was used properly
    if self.build_on_run and direct_call:
        raise RuntimeError("You used set_device with build_on_run=True "
                           "(the default option), which will automatically "
                           "build the simulation at the first encountered "
                           "run call - do not call device.build manually "
                           "in this case. If you want to call it manually, "
                           "e.g. because you have multiple run calls, use "
                           "set_device with build_on_run=False.")
    
    # Check if network was already running before (FIXME necessarry?)
    if self.did_run:
        raise RuntimeError("The network has already been built and run "
                           "before. To build several simulations in "
                           "the same script, call \"device.reinit()\" "
                           "and \"device.activate()\". Note that you "
                           "will have to set build options (e.g. the "
                           "directory) and defaultclock.dt again.")
    
    # Prepare working directory
    self.prepare_directory()
    
    # TODO Unique network object names necessary?
    # See: https://github.com/brian-team/brian2/blob/master/brian2/devices/cpp_standalone/device.py#L1238
    
    # Get dt in seconds without unit
    dt_ = self.defaultclock.dt_
    # Add 'dt' to 'lava_variables' to make it available to Lava
    self.lava_variables['_defaultclock_dt']['definition'] = f'np.array([{dt_}])'

    if hardware == 'CPU':
        # Render Lava templates
        for obj in self.lava_objects:
            process_rendered, process_model_rendered = self.render_templates(obj)
            self.logger.diagnostic(
                f"Compiling templates:\nProcess:\n{process_rendered}\nProcess Model:\n{process_model_rendered}"
            )
            # Write to file
            self.write_templates(process_rendered, process_model_rendered,obj.name)

        # Run Lava process
        self.run()
    else:
        raise NotImplementedError(f'Hardware {hardware} is not implemented (yet), the list of available hardware: {self.available_hardware}')


def render_templates(self, obj):
    """
    Renders Jinja templates based on Brian network objects that are used in the lava taplates.
    We call them `lava objects`.

    Parameters
    ----------
    obj : lava_object
        Lava related network object
    
    Returns
    -------
    process_rendered : `string`
        A rendered lava `process` template
    process_model_rendered : `string`
        A rendered lava `process model` template
    """
    
    # Extract variables and abstract code
    process_methods, process_model_methods = self.get_compiled_code(obj)
    
    # Log extracted lava code
    s = "Extracted process methods:\n"
    for item in process_methods:
        s += f'{item}\n'
    self.logger.diagnostic(s)

    s = "Extracted process model methods:\n"
    for item in process_model_methods:
        s += f'{item}\n'
    self.logger.diagnostic(s)

    # Get a list of ordered function calls to be implemented in
    # the 'run' and '__init__' function
    # TODO rename to 'process' and 'process_model' functions
    lava_init_function_calls, lava_multi_init, lava_run_function_calls = self.get_lava_function_calls(obj)

    # Get the port definitions for process and process model
    proc_ports, proc_model_ports = self.get_lava_ports_definitions(obj)
    
    # Get formatted variables for lava process
    proc_variables_init, proc_variables_lava = self.get_lava_proc_variables(obj, lava_init_function_calls)
    
    # Log extracted lava process variables
    s = "Extracted lava process variables:\n"
    for item in proc_variables_lava:
        s += f'{item}\n'
    self.logger.diagnostic(s)
    
    # Get formatted variables for lava process **model**
    lava_proc_model_variables = self.get_lava_proc_model_variables(obj)
    
    # Add the port initializations:
    proc_variables_lava = proc_ports + proc_variables_lava
    lava_proc_model_variables = proc_model_ports + lava_proc_model_variables

    # Log extracted lava process variables
    s = "Extracted lava process model variables:\n"
    for item in lava_proc_model_variables:
        s += f'{item}\n'
    self.logger.diagnostic(s)
    
    # Get jinja environment
    env = self.get_jinja_environment()

    # Load and render 'process'
    process_template = env.get_template('process.py.j2')
    process_rendered = process_template.render(
        variables_init=proc_variables_init,
        variables_lava=proc_variables_lava,
        multi_init_calls = lava_multi_init,
        init_methods=process_methods,
        required_imports=collect_required_imports(process_methods),
        name = obj.name
    )
    
    # Load and render 'process model'
    process_model_template = env.get_template('process_model.py.j2')
    process_model_rendered = process_model_template.render(
        methods=process_model_methods,
        run_functions=lava_run_function_calls,
        variables=lava_proc_model_variables,
        required_imports=collect_required_imports(process_model_methods),
        name = obj.name
    )
    
    return process_rendered, process_model_rendered


def get_jinja_environment(self):
    """
    Creates a Jinja environment.

    The environment contains a loader which includes a path to the templates.
    
    Returns
    -------
    env : Environment
        A jinja environment that contains a loader with a path to the Jinja template files
    """
    
    # Get path to templates
    template_path = os.path.join(self.package_root, 'templates')
    
    # Defined Jinja file system loader based on a path to the template files
    loader = FileSystemLoader(searchpath=template_path)
    
    # Return the environment, containing the file loader
    return Environment(
        loader=loader,
        trim_blocks=True,
        lstrip_blocks=True
    )


def get_compiled_code(self, obj):
    """
    Collects the compiled code for lava process and lava process model

    Parameters
    ----------
    obj : lava_object
        Lava related network object

    Returns
    -------
    process_methods : `string[]`
        Process methods to include into the `process`
    process_model_methods : `string[]`
        Process methods to include into the `process model`
    """
    
    # Define variables to collect lava code
    process_methods = ''
    process_model_methods = ''
        
    # Iterate over code objects
    for code_object in self.code_objects.values():
        if code_object.owner.name == obj.name:
            lava_code_tmp = None
            for block in ['before_run','run','after_run']:
                # Get compiled code for specific code object and block
                lava_code_tmp = code_object.compiled_code[block]
                
                # Add the code collected from the code objects to either
                # the Lava process or the Lava process model
                if lava_code_tmp is not None:
                    if code_object.template_name in self.init_template_functions:
                        process_methods += lava_code_tmp + '\n\n'
                    else:
                        process_model_methods += lava_code_tmp + '\n\n'

    return process_methods.splitlines(), process_model_methods.splitlines()


def collect_required_imports(abstract_code):
    """
    Search for functions in abstract code that require an import (e.g. random function)
    and return these imports as array.
    
    Parameters
    ----------
    abstract_code : 'string''
        The whole abstract code as string
    
    Returns
    -------
    required_imports : `string[]`
        Array of strings that contain required imports
    """
    
    # Define potential imports
    potential_imports = {
        'random': 'from random import random',
        'timestep': 'from brian2.core.functions import timestep',
        'LazyArange': 'from brian2.codegen.runtime.numpy_rt.numpy_rt import LazyArange',
        'ceil': 'from brian2.codegen.generators.numpy_generator import ceil_func as ceil',
        'floor': 'from brian2.codegen.generators.numpy_generator import floor_func as floor',
        'int': 'from brian2.codegen.generators.numpy_generator import int_func as int',
        'rand': 'from brian2.codegen.generators.numpy_generator import randn_func as rand',
        'randn': 'from brian2.codegen.generators.numpy_generator import randn_func as randn',
        'poisson': 'from brian2.codegen.generators.numpy_generator import poisson_func as poisson',
        'logical_not': 'from numpy import logical_not',
        'sign': 'from numpy import sign',
        'abs': 'from numpy import abs',
        'sqrt': 'from numpy import sqrt',
        'exp': 'from numpy import exp',
        'log': 'from numpy import log',
        'log10': 'from numpy import log10',
        'sin': 'from numpy import sin',
        'cos': 'from numpy import cos',
        'tan': 'from numpy import tan',
        'sinh': 'from numpy import sinh',
        'cosh': 'from numpy import cosh',
        'tanh': 'from numpy import tanh',
        'arcsin': 'from numpy import arcsin',
        'arccos': 'from numpy import arccos',
        'arctan': 'from numpy import arctan',
        
    }
    
    # Create empty array to collect required imports
    required_imports = []
    # Check if relevant function is in abstract code and if yes, add import
    for func, imp in potential_imports.items():
        for line in abstract_code:
            if f'{func}(' in line:
                required_imports.append(imp)
                # avoid multiple imports of the same function
                break
    
    return required_imports     

def get_lava_ports_definitions(self, obj):
    """
    TODO

    Parameters
    ----------
    obj : lava_object
        Lava related network object
    
    Returns
    -------
    proc_ports : `string[]`
        A list of code lines that define process ports
    proc_model_ports : `string[]`
        A list of code lines that define process model ports
    """

    proc_ports = []
    proc_model_ports = []
    # Use the information contained in the objects to format the input and output ports
    if isinstance(obj, NeuronGroup) or isinstance(obj,SpikeGeneratorGroup):
        # Add the spikes_out ports, NOTE that the name has to be compatible with the name used
        # in the SpikeMonitor.
        spike_port = obj.name + '_s_out'
        proc_ports.append(f"self.{spike_port} = OutPort(shape= ({obj.N},))")
        proc_model_ports.append(f"{spike_port}: PyOutPort = LavaPyType(PyOutPort.VEC_DENSE, bool, precision=1)")
        for var in self.lava_ports.values():
            if not obj.name == var['receiver']:
                continue
            portname = var['portname']
            proc_ports.append(f"self.{portname}_in = InPort(shape=(0,))",)
            port_type = 'float' if not 'idx' in portname else 'int,precision = 1'
            proc_model_ports.append(f"{portname}_in: PyInPort = LavaPyType(PyInPort.VEC_DENSE, {port_type})")   
            
    elif isinstance(obj,Synapses):
        # First receive the incoming spikes from the neurons
        for pathway in obj._pathways:
            # Note that in this case the port doesn't need the obj.name prefix since we don't read values from it!
            prepost = pathway.prepost
            proc_ports.append(f'self.s_in_{prepost} = InPort(shape=(0,))')
            proc_model_ports.append(f"s_in_{prepost}: PyInPort = LavaPyType(PyInPort.VEC_DENSE, bool, precision=1)")
        # Then make ports for synaptic transmission to neurons
        for var in self.lava_ports.values():
            for pathway in obj._pathways:
                if not var['pathway'] == pathway:
                    continue
                portname = var['portname']
                shape_var = self.get_array_name(obj.variables['_synaptic_pre'], prefix = 'self.init')
                proc_ports.append(f"self.{portname}_out = OutPort(shape = {shape_var}.shape)")
                port_type = 'float' if not 'idx' in portname else 'int,precision = 1'
                proc_model_ports.append(f"{portname}_out: PyOutPort = LavaPyType(PyOutPort.VEC_DENSE, {port_type})")

    # If there are aliases of the same variable, make sure to initialize them only once
    proc_ports = list(set(proc_ports))
    # Add a return just to make the code slightly cleaner
    proc_ports[-1] += '\n'
    proc_model_ports = list(set(proc_model_ports))
    proc_model_ports[-1] += '\n'

    return proc_ports, proc_model_ports


def get_lava_proc_variables(self, obj, lava_init_function_calls):
    """
    Takes variable name/value pairs and generates a list of variables
    for the lava process

    Parameters
    ----------
    obj : lava_object
        Lava related network object
    lava_init_function_calls : `string[]`
        A list of strings containing function calls for the process
    
    Returns
    -------
    formatted_variables_init : `string[]`
        A list of code lines that contain variable declarations for the `process`
    formatted_variables_lava : `string[]`
        A list of code lines that contain variable declarations for the `process model`
    """
    
    # Store formatted init variables for a Lava process
    # This contains all variables again, but initialized as plain numpy arrays
    formatted_variables_init = []
    formatted_variables_lava = []

    for name, var in self.lava_variables.items():
        if not var['owner'] == obj.name and not var['owner'] == obj.clock.name:
            continue
        elif var['owner'] == obj.name:
            init_var_name = f'self.init{name}'
            numpy_definition = var["definition"]

            # Statement for the definition of an array variable in Lava
            formatted_variables_init.append(f'{init_var_name} = {numpy_definition}')
            
            # Check if Brian provides us with an init function for the variable,
            # that contains instructions to set user-defined initial values
            init_func = None
            if name in lava_init_function_calls:
                init_func = lava_init_function_calls[name]
                exp = f'Var(shape={var["shape"]}, init={init_func})'
            # Otherwise init with init variable that contains a plain numpy definition
            else:
                # Here I look at the shape of the init variable instead, it's useful if at the beginning
                # the shape is set to (0,) in brian due to the fact that they will resize this array.
                # This is because the lava shape of a variable doesn't get update according to the shape of its
                # init value.
                exp = f'Var(shape={init_var_name}.shape, init={init_var_name})'
        # NOTE: implementing multiple clocks will require checking on each obj.clock.name
        elif var['owner'] == obj.clock.name:
            dt = obj.clock.dt_
            exp = f'Var(shape= (1,), init = np.array([{dt}]))'
        
        # Statement for the definition of an array variable in Lava
        formatted_variables_lava.append(f'self.{name} = {exp}')

    return formatted_variables_init, formatted_variables_lava


def get_lava_proc_model_variables(self, obj):
    """
    Takes variable name/value pairs and generates a list of variables for the lava process model.

    Parameters
    ----------
    obj : lava_object
        Lava related network object
    
    Returns
    -------
    formatted_variables : str[]
        A list of code lines that contain variable declarations
    """
    
    # Init variable to store formatted variables for a Lava process model
    formatted_variables = []
    
    # Then the variables themselves
    for name, var in self.lava_variables.items():
        if var['owner'] == obj.name or var['owner'] == obj.clock.name:
            # Check if array or not
            value_type_arr = 'np.ndarray' if var["size"] > 1 else var["type"]

            # Format the expression to what a Lava process expects
            exp = f'LavaPyType({value_type_arr}, {var["type"]})'

            # Statement for the definition of an array variable in Lava
            formatted_variables.append(f'{name}: {value_type_arr} = {exp}')

            # Statement for the definition of an actual variable
            #formatted_variables.append(f'{var["name"]}: {value_type_arr} = {exp}')
        
    return formatted_variables


def get_lava_function_calls(self, obj):
    """
    Given the code objects we return an ordered list of function calls that should
    happend within our code. The ordering should be made more customizable

    Parameters
    ----------
    obj : lava_object
        Lava related network object

    Returns
    -------
    init_calls : `string[]`
        A list of code that describes methods for the `process`
    run_calls : `string[]`
        A list of code that describes methods for the `process model`
    """
    init_calls_lists = {}
    init_calls = {}
    run_calls = []
    multi_var_init_calls = []

    # Collect code objects for process
    code_objects = [c_o for c_o in list(self.code_objects.values()) if c_o.owner.name == obj.name]

    # Iterate over all code blocks and code objects
    # NOTE: The after_run code blocks are not really used at any point yet. 
    # FIXME: I take them out for now, because their behavior should be implemented differently!
    for block in ['_before_run()', '_run()']:
        for code_obj in code_objects:
            # If the codeobject is not empty, assign function names to related lists
            if code_obj.compiled_code[block[1:-2]] is not None:
                function_name = f'self.{code_obj.name}{block}'
                # Currently added synapses_create_array to the init functions because I believe we should only call it once.
                #if 'group_variable_set_conditional' in code_obj.name or block == '_before_run()' or 'synapses_create_array' in code_obj.name: 
                # # TODO: Here we might have to make sure that the 'group_...' gets added to index 0
                #    init_calls.append(function_name)
                # These functions are added to the lava process and initialize variables
                if code_obj.template_name in self.init_template_functions:
                    # Get variable that will be initialized in process
                    var_name = self.init_variables[code_obj.name]
                    # TODO: this can for sure be done more cleanly, but for now it suffices, check synapses.py to see why this works
                    if isinstance(var_name,str):
                        # If variable was not already added to list, just add it    
                        if var_name not in init_calls_lists:
                            init_calls_lists[var_name] = [function_name]
                        # If variable was already added to list, append it to the existing item
                        else:
                            init_calls_lists[var_name].append(function_name)
                    else:
                        multi_var_init_calls.append(function_name)
                # These functions handle the simulation and are part of the lava process model
                else:
                    run_calls.append(function_name)
    run_calls = schedule_sort(run_calls, obj)

    # Postprocess init calls
    for key, values in init_calls_lists.items():
        # If we have only one init function for the variable, just take it
        if len(values) == 1:
            init_calls[key] = values[0]
        # If we have two or more init functions for the variable,
        # we need to construct a lambda function that calls all init functions
        else:
            # Starting string definition for lambda function that calls all functions
            st = '(lambda: ['
            suffix = '])()[-1]'
            # Iteratively add all function calls as string
            for i, v in enumerate(values):
                st += f'{v}'
                if (i+1) < len(values):
                    st += ', '
            # Add suffix and add to init calls
            st += suffix
            init_calls[key] = st

    # If in any of our code objects there's a time variable (probably always)
    # Add a line to update at each time step
    time_update_set = False
    timestep_update_set = False
    for code_obj in code_objects:
        code_obj_varnames = code_obj.variables.keys()
        if 't' in code_obj_varnames and not time_update_set:
            # TODO: this has to be made more generalizable to multiple clocks
            run_calls.append('self._defaultclock_t += self._defaultclock_dt')
            time_update_set = True
        if 't_in_timesteps' in code_obj_varnames and not timestep_update_set:
            run_calls.append('self._defaultclock_timestep += 1')
            timestep_update_set = True

    return init_calls, multi_var_init_calls, run_calls


def schedule_sort(func_list, obj):
    """
    TODO

    Parameters
    ----------
    obj : lava_object
        Lava related network object

    Returns
    -------
    ordered_list
        A list containing the schedule, i.e the order of exectutions for code objects
    """

    from itertools import chain
    from brian2 import CodeRunner
    schedule = {
        'start': [],
        'groups': [],
        'thresholds': [],
        'synapses': [],
        'resets': [],
        'end': []
    }
    for func_call in func_list:
        if not isinstance(obj,CodeRunner):
            code_runner = [item for item in obj.contained_objects if item.name in func_call]
        else:
            # The only supported object which is itself a CodeRunner is the SpikeGeneratorGroup
            assert type(obj) == SpikeGeneratorGroup
            code_runner = [obj]
        
        #If the function doesn't correspond to the contained objects then it must be
        # the activation_processing code object, which doesn't have a corresponding CodeRunner object.
        # NOTE: This might be changed in future updates
        if not len(code_runner):
            # We want to receive the activations and update the neuron at the start of the timestep.
            # NOTE: The various run_ functions from lava might prove useful here in the future.
            assert 'activation_processing' in func_call
            schedule['synapses'].insert(0, func_call)
            continue

        if not len(code_runner) == 1:
            raise ValueError(f"""More than one CodeRunner corresponding to the same code_object. 
            Try restarting the simulation. If the bug persists please report it to us.
            CodeRunners: {code_runner}""")
        code_runner = code_runner[0]
        schedule[code_runner.when].insert(code_runner.order,func_call)
    ordered_list = []
    for when in schedule:
        ordered_list = list(chain(ordered_list, schedule[when]))

    return ordered_list



import numpy as np

# from brian2.core.base import BrianObjectException
# from brian2.core.preferences import prefs, BrianPreference
from brian2.core.variables import DynamicArrayVariable, ArrayVariable, AuxiliaryVariable, Subexpression
from brian2.core.functions import Function

from brian2.codegen.codeobject import CodeObject, constant_or_scalar, check_compiler_kwds
from brian2.codegen.targets import codegen_targets

from brian2lava.codegen.lava_generator import LavaCodeGenerator
from brian2lava.codegen.templater import Templater


class LavaCodeObject(CodeObject):
    """
    Class of code objects that generate Lava compatible code
    """


    templater = Templater('brian2lava.codegen','.py_', env_globals={'constant_or_scalar': constant_or_scalar})
    generator_class = LavaCodeGenerator
    class_name = 'lava'


    def __init__(
            self, owner, code, variables, variable_indices,
            template_name, template_source, compiler_kwds,
            name='lava_code_object*'
        ):
        check_compiler_kwds(compiler_kwds, [], 'lava')

        from brian2.devices.device import get_device
        self.device = get_device()
        self.namespace = {
            '_owner': owner,
            'logical_not': np.logical_not  # TODO: This should maybe go somewhere else
        }
        CodeObject.__init__(
            self, owner, code, variables, variable_indices,
            template_name, template_source,
            compiler_kwds=compiler_kwds, name=name
        )
        self.variables_to_namespace()


    @classmethod
    def is_available(cls):
        """
        Checks if the given backend is available

        Parameters
        ----------
        cls
            A CodeObject derived class

        Returns
        -------
        `bool`
            Indicates if the given backend is avialable or not
        """

        # TODO
        # For all hardwares perhaps check if Lava is installed?
        # For Loihi, we need to check if Loihi is actually available
        # For now, just return true
        return True


    def variables_to_namespace(self):
        """
        Adds variables to the Brian namespace.

        Notes
        -----
        Variables can refer to values that are either constant (e.g. dt) or change every timestep (e.g. t).
        We add the values of the constant variables here and add the names of non-constant variables to a list.
        """

        # A list containing tuples of name and a function giving the value
        self.nonconstant_values = []

        for name, var in self.variables.items():
            if isinstance(var, (AuxiliaryVariable, Subexpression)):
                continue

            try:
                if not hasattr(var, 'get_value'):
                    raise TypeError()
                value = var.get_value()
            except TypeError:
                # Either a dummy Variable without a value or a Function object
                if isinstance(var, Function):
                    impl = var.implementations[self.__class__].get_code(self.owner)
                    self.namespace[name] = impl
                else:
                    self.namespace[name] = var
                continue

            if isinstance(var, ArrayVariable):
                self.namespace[self.generator_class.get_array_name(var)] = value
                if var.scalar and var.constant:
                    self.namespace[name] = value[0]
            else:
                self.namespace[name] = value

            if isinstance(var, DynamicArrayVariable):
                dyn_array_name = self.generator_class.get_array_name(var, access_data=False)
                self.namespace[dyn_array_name] = self.device.get_value(var, access_data=False)

            # Also provide the Variable object itself in the namespace (can be
            # necessary for resize operations, for example)
            self.namespace[f"_var_{name}"] = var

            # There is one type of objects that we have to inject into the
            # namespace with their current value at each time step: dynamic
            # arrays that change in size during runs (i.e. not synapses but
            # e.g. the structures used in monitors)
            if (isinstance(var, DynamicArrayVariable) and var.needs_reference_update):
                self.nonconstant_values.append((
                    self.generator_class.get_array_name(var, self.variables),
                    var.get_value
                ))


    def update_namespace(self):
        """
        Updates variables from the code object in the Brian namespace.
        """

        # Update the values of the non-constant values in the namespace
        for name, func in self.nonconstant_values:
            self.namespace[name] = func()


    def compile_block(self, block):
        """
        Compiles a block of code.

        Parameters
        ----------
        block
            A block of code

        Returns
        -------
        `str`
            Compiled code
        """

        code = getattr(self.code, block, '').strip()
        if not code or 'EMPTY_CODE_BLOCK' in code:
            return None
        # TODO What does python's compile return?
        return code


    def run(self):
        """
        Runs the code.

        Notes
        -----
        The execution of compiled code is deligated to the device.
        """

        pass


    def run_block(self, block):
        """
        Runs a block of code.

        Parameters
        ----------
        block
            A block of code

        Notes
        -----
        The execution of compiled code is deligated to the device.
        """

        pass


codegen_targets.add(LavaCodeObject)



import os
from jinja2 import ChoiceLoader, Environment, StrictUndefined, FileSystemLoader
from brian2.codegen.templates import autoindent, variables_to_array_names, LazyTemplateLoader


class Templater:
	"""
	This class replaces `from brian2.codegen.templates import Templater`.
	This is required, since Jinja PackageLoader has a problem with packaging,
	the PackageLoader is here replaced by the FileSystemLoader.

	Class to load and return all the templates a `CodeObject` defines.

	Parameters
	----------
	package_name : str, tuple of str
		The package where the templates are saved. If this is a tuple then each template will be searched in order
		starting from the first package in the tuple until the template is found. This allows for derived templates
		to be used. See also `~Templater.derive`.
	extension : str
		The file extension (e.g. ``.pyx``) used for the templates.
	env_globals : dict (optional)
		A dictionary of global values accessible by the templates. Can be used for providing utility functions.
		In all cases, the filter 'autoindent' is available (see existing templates for example usage).
	templates_dir : str, tuple of str, optional
		The name of the directory containing the templates. Defaults to ``'templates'``.
	Notes
	-----
	Templates are accessed using ``templater.template_base_name`` (the base name is without the file extension).
	This returns a `CodeObjectTemplate`.
	"""

	def __init__(
		self, package_name, extension, env_globals=None, templates_dir="codegen/templates"
	):
		if isinstance(package_name, str):
			package_name = (package_name,)
		if isinstance(templates_dir, str):
			templates_dir = (templates_dir,)

		# Define project root
		project_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
		
		templateLoader = []
		for t_dir in templates_dir:
			# Get path to templates
			template_path = os.path.join(project_root, t_dir)
			# Define file system loader
			fsl = FileSystemLoader(template_path)
			templateLoader.append(fsl)

		loader = ChoiceLoader(templateLoader)

		self.env = Environment(
			loader=loader,
			trim_blocks=True,
			lstrip_blocks=True,
			undefined=StrictUndefined,
		)
		self.env.globals["autoindent"] = autoindent
		self.env.filters["autoindent"] = autoindent
		self.env.filters["variables_to_array_names"] = variables_to_array_names
		if env_globals is not None:
			self.env.globals.update(env_globals)
		else:
			env_globals = {}
		self.env_globals = env_globals
		self.package_names = package_name
		self.templates_dir = templates_dir
		self.extension = extension
		self.templates = LazyTemplateLoader(self.env, extension)


	def __getattr__(self, item):
		return self.templates.get_template(item)


	def derive(
		self, package_name, extension=None, env_globals=None, templates_dir="templates"
	):
		"""
		Return a new Templater derived from this one, where the new package name and globals overwrite the old.
		"""
		
		if extension is None:
			extension = self.extension
		if isinstance(package_name, str):
			package_name = (package_name,)
		if env_globals is None:
			env_globals = {}
		if isinstance(templates_dir, str):
			templates_dir = (templates_dir,)
		package_name = package_name + self.package_names
		templates_dir = templates_dir + self.templates_dir
		new_env_globals = self.env_globals.copy()
		new_env_globals.update(**env_globals)
		return Templater(
			package_name,
			extension=extension,
			env_globals=new_env_globals,
			templates_dir=templates_dir,
		)



import numpy as np

from brian2.memory.dynamicarray import DynamicArray, DynamicArray1D
from brian2.core.variables import ArrayVariable, DynamicArrayVariable

from brian2.monitors.spikemonitor import SpikeMonitor
from brian2.monitors.statemonitor import StateMonitor

def get_value(self, var, access_data=True):
    """
    Get a value from an array.
    Returning a value from the device arrays depends on the type of array,
    which can either be a static array (ArrayVariable) or a dynamic array (DynamicArrayVariable).

    Parameters
    ----------
    var : `ArrayVariable`
        The array to get
    access_data : `bool`
        A flag that indicates if it is intended to access only the data of the dynamic array (True)
        or the whole dynamic array (False)

    Returns
    -------
    `any`
        Values of the array variable as list
    """

    # Log that a value was requested from arrays
    self.logger.diagnostic(f'get_value {var.name}')

    return self.arrays[var]


def get_var_by_lava_name(self, lava_var_name):
    """
    Get an array variable from array by lava variable name

    Parameters
    ----------
    lava_var_name : `string`
        The name of the lava variable

    Returns
    -------
    `ArrayVariable`
        An array variable
    """

    # Iterate over all arrays
    for var, values in self.arrays.items():
        # Check if variable is not DynamicArrayVariable, since we don't want to get monitor variables
        is_no_monitor = not isinstance(var, DynamicArrayVariable)
        # Check if name of variable matches given name
        is_name_match = self.get_lava_var_name(var) == lava_var_name

        # If both conditions match, break loop and return variable
        if is_no_monitor and is_name_match:
            return var


def get_lava_var_name(self, var):
    """
    Get a lava variable name based on an array variable.

    Parameters
    ----------
    var : `ArrayVariable`
        An array variable

    Returns
    -------
    `string`
        The corresponding lava variable name

    Notes
    -----
    TODO This can possibly be hamrmonized with `get_array_name`.
    """

    if isinstance(var.owner, StateMonitor):
        source_name = 'defaultclock' if var.name == 't' else var.owner.source.name
        #source_name = var.owner.name
        return f'_{source_name}_{var.name}'
    elif isinstance(var.owner, SpikeMonitor):
        lava_var_name = ''
        if var.name == 't': lava_var_name = '_defaultclock_t'
        if var.name == 'i': lava_var_name = var.owner.source.name +'_s_out'
        return lava_var_name
    else:
        return f'_{var.owner.name}_{var.name}'


def get_array_name(self, var, access_data=True, prefix='self.'):
    """
    Gets the name of an array variable.

    Parameters
    ----------
    var : `ArrayVariable`
        The array to get.
    access_data : `bool`
        A flag that indicates if it is intended to access only the data of the dynamic array (True)
        or the whole dynamic array (False)
    prefix : `string`
        A string that is added as a prefix to the array name
        Default is 'self.', in case of 'None', no prefix is added

    Returns
    -------
    `string`
        The corresponding variable name as it is used in Brian

    Notes
    -----
    TODO This can possibly be hamrmonized with `get_lava_var_name`.
    """
    
    # The name of the array is part of the owner attribute
    # The owner is a `Nameable`, e.g. `NeuronGroup` or `Synapses`
    # If no owner name is available, 'temporary' is assigned
    owner_name = getattr(var.owner, 'name', 'temporary')

    # Redefine prefix to empty string if it was set to 'None'
    if prefix is None:
        prefix = ''
    
    return f'{prefix}_{owner_name}_{var.name}'


def get_monitor_type_name(owner):
    """
    Get monitor type name, i.e. spike monitor or state monitor

    Parameters
    ----------
    owner : `StateMonitor` or `SpikeMonitor`
        The owner of a group

    Returns
    -------
    `string`
        The corresponding monitor type as string
    """

    # Init monitor type variables
    monitor_type_name = ''

    # Check instance of owner and define types
    if isinstance(owner, StateMonitor):
        monitor_type_name = 'state'
    elif isinstance(owner, SpikeMonitor):
        monitor_type_name = 'spike'
    else:
        raise Exception('Unknown owner instance. Owner instance has to be StateMonitor or SpikeMonitor.')

    return monitor_type_name


def add_array(self, var):
    """
    Add a variable array to the `arrays` list of the device.
    It can either be added directly or as a `DynamicArrayVariable` object.
    The `DynamicArrayVariable` can dynamically be extended (in contrast to a static array).

    We separate between monitors and all other owner types of the variable to add.
    Monitors are added to the `lava_monitors` list. All other variable types are added to `lava_variable_names`.

    Parameters
    ----------
    var : `ArrayVariable`
        The array variable to add
    """

    # NOTE only for Loihi hardware, on CPU this is not necessary
    # Only add array if owner is of class SpikeMonitor or StateMonitor
    #if not isinstance(var.owner, (SpikeMonitor, StateMonitor)):
    #    return

    # Log that a value was added to arrays
    self.logger.diagnostic(f'add_array {var.name}')

    # Create a static numpy array
    arr = np.empty(var.size, dtype=var.dtype)

    # Add array to device arrays
    self.arrays[var] = arr

    if isinstance(var.owner, (SpikeMonitor, StateMonitor)):
        # NOTE Currently only dynamic array variables of a monitor (like v, t, etc.) are added
        #      Constant values like N or __indices are currently ignored
        if isinstance(var, DynamicArrayVariable):

            # Get monitor type name from owner ('state' or 'spike')
            monitor_type_name = get_monitor_type_name(var.owner)

            # Define monitor name and lava variable name
            monitor_name = f'_{monitor_type_name}_{var.owner.name}_{var.name}'
            lava_var_name = self.get_lava_var_name(var)

            # Collect lava variable names that shall be monitored by lava
            # NOTE Only add variable name if it's not already in the list
            #      Variable names can occure in multiple monitors, e.g. time
            if lava_var_name not in self.lava_variables_to_monitor:
                self.lava_variables_to_monitor.append(lava_var_name)
            # Define monitor
            self.lava_monitors[monitor_name] = {
                'name' : monitor_name, # The name of this monitor, mainly for debugging
                'source': var.owner.source.name,
                'var': var,  # Brian variable
                'lava_var_name': lava_var_name,  # The variable name used in Lava
                'type': self.monitor_types[monitor_type_name],  # The monitor type ('state' or 'spike')
                'lava_monitor': None  # The Lava monitor, instance is added later during 'run'
            }
    else:
        # Add the definition of a numpy array as string for lava
        var_definition = f'np.empty({var.size}, dtype=np.{arr.dtype.name})'

        # TODO this can be done better probably
        dtype = arr.dtype.name        
        value_type = dtype.replace('32', '')
        value_type = value_type.replace('64', '')

        # TODO is the key unique?
        # See also: https://github.com/brian-team/brian2/pull/304
        name = f'_{var.owner.name}_{var.name}'
        self.lava_variables[name] = {
            'name': var.name,
            'owner': var.owner.name,
            'definition': var_definition,
            'size': var.size,
            'shape': np.shape(arr),
            'type': value_type,
            'dtype': arr.dtype.name
        }

        # NOTE information from this dict can also be obtained from self.lava_variables
        #      directly, but it's a bit complicated, for now we can leave it as a kind of cache
        self.lava_variable_names[var.name] = name


def init_with_zeros(self, var, dtype):
    """
    Initialize an array with zeros and adds it to the `arrays` list.

    Parameters
    ----------
    var : `ArrayVariable`
        The array variable to initialize with zeros
    dtype : `dtype`
        The data type to use for the array
    """

    # Redefine variable definition for Lava variables
    name = f'_{var.owner.name}_{var.name}'
    if name in self.lava_variables.keys():
        lv = self.lava_variables[name]
        lv['definition'] = f'np.zeros({lv["size"]}, dtype=np.{lv["dtype"]})'
    
    # Log that an empty array was initialized
    self.logger.diagnostic(f'init_with_zeros {var.name}')
    
    self.arrays[var][:] = 0


def init_with_arange(self, var, start, dtype):
    """
    Initializes an array using the numpy arange function and adds it to the `arrays` list.
    The `start` value defines the start of the range, the length is given by the length of the `var` array.
    
    Parameters
    ----------
    var : `ArrayVariable`
        The array to initialize is based on the length of this `var` array
    start : `int`
        Start value of the range
    dtype : `dtype`
        The data type to use for the array
    """

    # Redefine variable definition for Lava variables
    name = f'_{var.owner.name}_{var.name}'
    if name in self.lava_variables.keys():
        lv = self.lava_variables[name]
        lv['definition'] = f'np.arange({start}, {lv["size"]+start}, dtype=np.{lv["dtype"]})'
    
    # Log that an array was created based on numpy arange
    self.logger.diagnostic(f'init_with_arange, arange from {start} to {var.get_len()+start}')
    
    self.arrays[var][:] = np.arange(start, stop=var.get_len()+start, dtype=dtype)


def fill_with_array(self, var, arr):
    """
    Fill array vatiable `var` with the values given in an array `arr` and add it to the `arrays` list.
    
    Parameters
    ----------
    var : `ArrayVariable`
        The array variable to fill
    arr : `ndarray`
        The array values that should be copied to `var`
    """

    arr = np.asarray(arr)
    # Redefine variable definition for Lava variables
    name = f'_{var.owner.name}_{var.name}'
    if name in self.lava_variables.keys():
        lv = self.lava_variables[name]

        # Check if 'arr' is given as scalar
        is_scalar = not bool(len(np.shape(arr)))
        
        # If 'arr' is scalar and size is 1, add a simple array
        if is_scalar and lv['size'] == 1:
            lv['definition'] = f'np.array([{arr}], dtype=np.{lv["dtype"]})'
        # If 'arr' is scalar and size > 1, add an array that repeats the value accordingly
        elif is_scalar and lv['size'] > 1:
            lv['definition'] = f'np.array(np.repeat({arr}, {lv["size"]}), dtype=np.{lv["dtype"]})'
        # If 'arr' is actually an array, it's just transformed to a string definition
        else:
            arr_str = np.array2string(np.array(arr), separator=', ')
            lv['definition'] = f'np.array({arr_str}, dtype=np.{lv["dtype"]})'

    # Log that an array was filled with given values
    self.logger.diagnostic(f'fill_with_array, add {arr} to {var.name}')
    
    # Set array value
    self.arrays[var][:] = arr

def resize(self, var, new_size):
    """
    Method called most times a DynamicArray variable is created. Updates the size of a DynamicArray.
    """
    # First resize it in our array-cache
    self.arrays[var].resize(new_size)

    # Change the size of the variable in our variable dictionary
    # We also make sure to update the default definition in case the variable
    # is not initialized with a 'fill_with' method afterwards (in most cases it is).
    # This is probably redundant but ensures no bugs come up later on.
    name = f'_{var.owner.name}_{var.name}'
    if name in self.lava_variables.keys():
        lv = self.lava_variables[name]
        lv["size"] = new_size
        if 'np.empty' in lv["definition"]:
            lv["definition"] = f'np.empty({lv["size"]}, dtype=np.{lv["dtype"]})'


    




import jinja2
import os
import numpy as np

#from types import SimpleNamespace
from weakref import WeakKeyDictionary

# Import Brian2 modules
from brian2.devices.device import all_devices, Device
from brian2.utils.logger import get_logger

__all__ = ['LavaDevice']


class LavaDevice(Device):
    """
    The Lava device combines most of the functionality. It activates the device,
    transforms all variables (to and from Lava), builds and executes the generated Lava code.

    The generation of the Lava Python code, based on the abstract code from Brian,
    is performed mostly from the `LavaCodeGenerator` and partly from `LavaCodeObject`.
    Also the `code_object` method within this `LavaDevice` takes part in generating Lava code.

    The LavaDevice is split into several files. The import statements below the init method
    include the code of the other files into the LavaDevice.
    """


    def __init__(self):
        super(LavaDevice, self).__init__()
        # Make logger available to all methods in this class
        self.logger = get_logger('brian2.devices.lava')

        # Store root directory of this package
        self.package_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

        # The project dir is fixed in the Lava device
        # This is important, since a reinit otherweise don't know which directory to clean
        self.project_dir = 'lava_workspace'

        # FIXME Defines the network schedule, if None, the user can choose the schedule
        self.network_schedule = None
        # NOTE Brian2Loihi used: ['start', 'synapses', 'groups', 'thresholds',  'resets', 'end']

        # Random number buffer
        self.randn_buffer_index = np.zeros(1, dtype=np.int32)
        self.rand_buffer_index = np.zeros(1, dtype=np.int32)

        # Predefines a default clock, it is set in the `activate` method
        self.defaultclock = None

        # Define an empty set to store clocks
        self.clocks = set([])

        # Stores weak references from a `Variable` objects to the containing value(s)
        # Methods to handle the arrays are provided in `arrays.py`
        self.arrays = {}  # FIXME is was originally a WeakKeyDictionary()

        # Define empty dicts to store code objects and abstract code
        self.code_objects = {}

        # Brian network objects
        self.net_objects = set()

        # Store Lava variables
        # Key: variable name, Value: variable definition (e.g. 'np.empty(...)')
        self.N = None
        self.lava_variables = {}
        self.lava_variable_names = {}
        
        # Store the ports required to connect processes
        self.lava_ports = {}

        # Brian template functions that belong to the process,
        # instead of the process model, e.g. 'group_variable_set'
        self.init_template_functions = [
            'group_variable_set',
            'synapses_create_generator',
            'synapses_create_array'
        ]
        
        # Variables that represent our current feature support
        from brian2 import SpikeGeneratorGroup, NeuronGroup, Synapses
        self.supported_processes = [
            NeuronGroup,
            Synapses,
            SpikeGeneratorGroup
        ]
        from brian2 import PoissonInput,PopulationRateMonitor,SpatialNeuron, PoissonGroup
        self.unsupported_processes = [
            PoissonInput,
            PopulationRateMonitor,
            SpatialNeuron,
            PoissonGroup
        ]

        self.available_hardware = ["CPU"]

        # Stores names of indices that are used to initialize variables
        # e.g. _group_idx_1_v, _group_idx_2_ge, etc.
        self.set_variable_index_names = []

        # Last method name for setting variables
        self.last_set_variable_method_name = ''

        # Maps process *method name* (for initializing variables) with
        # *variable name* (variable to initialize)
        self.init_variables = {}

        # Store Lava monitors
        # NOTE Some variables shall be probed only once by Lava, but are required
        #      for different Brian monitors, e.g. time
        self.lava_variables_to_monitor = []
        self.lava_monitors = {}
        #self.brian_monitors = {}

        self.monitor_types = {'state': 0, 'spike': 1}  # SimpleNamespace(**{ 'state': 0, 'spike': 1 })
        
        # Init a variable to store the Lava process
        self.process = None

        # Stores if the network did run
        self.did_run = False  # FIXME necessary?: Yes? since otherwise the user
        # can use multiple run calls but if he does it the objects and network
        # might have changed but the compiled code for lava is still the old
        # one. He needs to do a device.reini(), device.activat() to reinit the
        # whole build process again. WIth this flag here we chan check this.

        # Define constants
        #C = SimpleNamespace(**{
        #    CPU: 'CPU',
        #    LOIHI1: 'Loihi1',
        #    LOIHI2: 'Loihi2'
        #})

        # Add super to self to make it available in sub-files
        self.super = super(LavaDevice, self)

    
    # Device activation
    from .activate import (
        activate, seed, reinit
    )

    # Build the network
    from .build import (
        build, render_templates, get_compiled_code, get_jinja_environment,
        get_lava_proc_variables, get_lava_proc_model_variables,
        get_lava_function_calls, get_lava_ports_definitions
    )

    # Writer for writing rendered templates
    from .writer import (
        prepare_directory, write_templates
    )

    # Define code objects
    from .codeobject import (
        code_object,
        determine_lava_ports
    )

    # Run the network
    from .run import (
        network_run, run, set_brian_monitor_values, init_lava_monitors, connect_lava_ports,select_root_processes
    )

    # Handle storage
    from .arrays import (
        add_array, get_value, get_array_name, init_with_zeros,
        init_with_arange, fill_with_array, get_var_by_lava_name,
        get_lava_var_name, resize
    )

    # Handle setter and getter for variables
    from .variables import (
        variableview_set_with_index_array,
        variableview_set_with_expression,
        variableview_set_with_expression_conditional
    )

    # Handle synapses
    from .synapses import (
        synapses_connect,
        spike_queue,
        add_lava_synapses_generator
    )


# Add lava device to all_devices and make it with this available to Brian2
lava_device = LavaDevice()
all_devices['lava'] = lava_device



import numpy as np
import os
import shutil

from brian2.codegen.targets import *
from brian2.core.clocks import Clock
from brian2.core.preferences import prefs
from brian2.units import ms

from brian2lava.codegen.codeobject import LavaCodeObject


def activate(self, **kwargs):
    """
    Activates Brian2Lava device.

    The method adds the `LavaCodeObject` as code generation target
    """
    # Log that activate method was called
    self.logger.debug("Activating Lava device.")

    # Use Lava code object
    codegen_targets.add(LavaCodeObject)

    # Set codegen targets to 'lava'
    prefs.codegen.target = 'lava'
    prefs.codegen.string_expression_target = 'lava'

    # Log used device and code object
    self.logger.debug(f'Using code object class: {self.code_object_class().__name__} with Device: {self}')

    # Call parent activate function
    self.super.activate(**kwargs)


def seed(self, seed=None):
    """
    Set the seed for the random number generator.

    Parameters
    ----------
    seed : int, optional
        The seed value for the random number generator, or ``None`` (the default) to set a random seed.
    """
    np.random.seed(seed)
    self.rand_buffer_index[:] = 0
    self.randn_buffer_index[:] = 0


def reinit(self):
    """
    Reinitializes the device, which is necessary if multiple `run()` calls
    are performed within a single script.

    *   Initialize device and call parent's `reinit`
    *   Reset `did_run` flag 
    *   Set network schedule and 'build_on_run' flag to previously chosen values
    *   Clean workspace directory
    """
    # Store network schedule and 'build_on_run' flag
    tmp_network_schedule = self.network_schedule
    tmp_build_on_run = self.build_on_run

    # Initialize the device
    self.__init__()

    # Call reinit from super device
    self.super.reinit()  # necessary?

    # Resets the 'did_run' flag
    self.did_run = False

    # Set network schedule and 'build_on_run' flag to previously chosen values
    self.network_schedule = tmp_network_schedule
    self.build_on_run = tmp_build_on_run

    # Clean workspace: Get absolute workspace directory and remove recursively
    abs_project_dir = os.path.abspath(os.path.join(os.getcwd(), self.project_dir))
    shutil.rmtree(abs_project_dir)



import os
import tempfile

# Import Brian2 modules
from brian2.utils.filetools import ensure_directory

# Import Brian2Lava modules
from brian2lava.writer.py import PyWriter


def prepare_directory(self):
    """
    Prepare the directory and make sure a given directory exists
    """
        
    # Create project directories (if not exists)
    ensure_directory(self.project_dir)
    #for d in ['code_objects', 'results', 'static_arrays']:
    #    ensure_directory(os.path.join(directory, d))

    # Log directory path
    self.logger.debug(
        "Writing Lava project to directory " + os.path.normpath(self.project_dir)
    )


def write_templates(self, process_rendered, process_model_rendered, name):
    """
    Write the rendered templates to files
    
    Parameters
    ----------
    process_rendered : `str`
        The rendered Lava process template as string
    process_model_rendered : `str`
        The rendered Lava process model template as string
    name : `str`
        Name of the network object (e.g. 'neurongroup_0')
    """
    
    # Instantiate python file writer
    writer = PyWriter(self.project_dir)
    
    # Write 'process' to working directory
    writer.write(f'{name}_process.py', process_rendered)
    
    # Log that process.py was stored
    self.logger.diagnostic(f'{name}_process.py stored in working directory.')
    
    # Write 'process model' to working directory
    writer.write(f'{name}_process_model.py', process_model_rendered)
    
    # Log that process_model.py was stored
    self.logger.diagnostic(f'{name}_process_model.py stored in working directory.')



import os
import importlib
import numpy as np

from brian2.core.clocks import Clock
from brian2.core.namespace import get_local_namespace
from brian2.core.network import TextReport, _get_all_objects
from brian2.groups.neurongroup import NeuronGroup
from brian2.synapses.synapses import Synapses
from brian2.units import second

from lava.magma.core.run_configs import Loihi1SimCfg
from lava.magma.core.run_conditions import RunSteps
from lava.proc.monitor.process import Monitor

from pprint import pprint


def network_run(
        self,
        net,
        duration,
        report=None,
        report_period=10 * second,
        namespace=None,
        profile=False,
        level=0,
        **kwds
    ):
    """
    Performs preparations and checks and finally calls the `build` method from the device.

    Parameters
    ----------
    TODO

    Notes
    -----
    This is called very early in the process and overwrites the `run()` method from `brian2.core.Network.run()`.
    """
    # Before doing anything check that the objects used by the user 
    # are supported by lava.
    check_for_brian2lava_support(net)
    
    # Store duration in device
    self.duration = duration
    
    # Log that the network run method was called
    self.logger.diagnostic("Network run is executing.")

    # Add the network from the current run call to the set of networks within this device
    #self.networks.add(net)
    
    # If keyword arguments are given, notify the user that these arguments are not used in Brian2Lava
    if kwds:
        self.logger.warn(
            'Unsupported keyword argument(s) provided for run: {}'.format(', '.join(kwds.keys()))
        )
        
    # FIXME Show an error if user enabled profiling, since it is not supported
    if profile is True:
        raise NotImplementedError('Brian2Lava does not yet support detailed profiling.')
    
    # FIXME Set clocks
    net._clocks = {obj.clock for obj in net.sorted_objects}
    t_end = net.t+duration
    for clock in net._clocks:
        clock.set_interval(net.t, t_end)
    
    # Get the local namespace, if no namespace was given
    if namespace is None:
        namespace = get_local_namespace(level=level+2)

    # Call before_run with namespace
    net.before_run(namespace)
    
    
    # Get all synapses objects
    #self.synapses |= {s for s in net.objects if isinstance(s, Synapses)}
    
    # Update device clocks by network clocks
    self.clocks.update(net._clocks)
    if len(self.clocks) > 1:
        raise NotImplementedError("Currently multiple clocks are not supported by brian2lava.")
    
    # Set current time to end time (FIXME why? and whyt is the difference between self.t and self.t_?)
    # NOTE by carlo: self.t is with unit, self.t_ is without unit
    net.t_ = float(t_end)

    # Gets the number of neurons for the neuron group
    # FIXME this currenty assumes only *one* NeuronGroup
    #       If multiple neuron groups are present, this needs to be changed
    for obj in net.objects:
        if isinstance(obj, NeuronGroup):
            self.N = obj._N
    
    # FIXME Taken from CPPStandaloneDevice, but unclear what it means
    # In the CPP device it is noted that this is a hack
    # https://github.com/brian-team/brian2/blob/master/brian2/devices/cpp_standalone/device.py#L1404
    for clock in self.clocks:
        if clock.name == 'clock':
            clock._name = '_clock'
    
    # Collect code objects from network objects
    # Right now these lines don't really do anything since we're never referencing the 
    # variable code_objects again in this method
    code_objects = []
    self.lava_objects = []
    for obj in net.sorted_objects:
        # FIXME: Is this required? What does obj.active actually do?
        if obj.active:
            # These are (currently) the only BrianObjects that generate their own process
            # NOTE: This must be changed when implementing other Brian classes
            if any([isinstance(obj,supp_obj) for supp_obj in self.supported_processes]):
                self.lava_objects.append(obj)
            for codeobj in obj._code_objects:
                code_objects.append((obj.clock, codeobj))

    # FIXME This may require an update, similar to how CPPStandalone handles the report
    # https://github.com/brian-team/brian2/blob/master/brian2/devices/cpp_standalone/device.py#L1465
    #if report is not None:
    #    report_period = float(report_period)
    #    next_report_time = start_time + report_period
    #    if report == 'text' or report == 'stdout':
    #        report_callback = TextReport(sys.stdout)
    #    elif report == 'stderr':
    #        report_callback = TextReport(sys.stderr)
    #    elif isinstance(report, str):
    #        raise ValueError(f'Do not know how to handle report argument "{report}".')
    #    elif callable(report):
    #        report_callback = report
    #    else:
    #        raise TypeError(f"Do not know how to handle report argument, "
    #                        f"it has to be one of 'text', 'stdout', "
    #                        f"'stderr', or a callable function/object, "
    #                        f"but it is of type {type(report)}")
    #    report_callback(0*second, 0.0, t_start, duration)

    
    # TODO? At least in the CPPStandaloneDevice, there is some more code here,
    # that seems to generate some basic code lines ...

    #for a in self.arrays:
    #    if (a.name == 't') and isinstance(a.owner, Clock):
    #        print(a.get_value())
    
    # Call network after_run method
    net.after_run()
    
    # Call build method
    if self.build_on_run:
        if self.did_run:  # FIXME necessary?
            raise RuntimeError("The network has already been built and run "
                               "before. Use set_device with "
                               "build_on_run=False and an explicit "
                               "device.build call to use multiple run "
                               "statements with this device.")
        self.build(direct_call=False, **self.build_options)

def check_for_brian2lava_support(net):
    """
    Checks that the objects defined by the user are supported by the lava device.
    If not, it throws a NotImplementedError
    Parameters
    ----------
    net : brian2.network.Network

    raises : NotImplementedError
            If any of the objects in the Network are not currently supported by brian2lava
    """
    from brian2 import get_device
    device = get_device()
    # Raise an error if the user is trying to implement unsupported objects
    objects = []
    for obj in net.sorted_objects:
        if any([isinstance(obj,unsupp_obj) for unsupp_obj in device.unsupported_processes]):
            obj_type = type(obj).__name__
            objects.append(obj_type)
        
    if len(objects):
        objects_repr = '\n\t\t'.join(objects)
        msg = f"""
        The following objects are not supported by brian2lava, yet:
                {objects_repr}
        You can expect them in future releases. You can also ask for features 
        on the official brian2lava repo on GitLab:
        https://gitlab.com/brian2lava/brian2lava/-/tree/main
        """
        raise NotImplementedError(msg)
        


def compile_templates(directory, names):
    """
    Compiles the rendered templates and returns the obtained process model

    Parameters
    ----------
    directory : `string`
        The project directory, necessary to find files to compile.
    names : network object names
        Names of network objects that are used for the Lava templates
    
    Returns
    -------
    compiled_processes: `Process`
        The compiled Lava process
    """
    # Execute Lava process and Lava process model
    # Add all outcomes to global scope
    compiled_processes = {}
    for name in names:
        exec(open(f'{directory}/{name}_process.py').read(), globals())
        exec(open(f'{directory}/{name}_process_model.py').read(), globals())
    
        exec(f"compiled_processes['{name}'] = {name}_P(name = '{name}')")
    
    # Get instance of the process
    # NOTE Linting assumes that 'Process' is not defined it does not
    #      understand that it comes from the exec above
    return compiled_processes  # noqa: F821


def run(self):
    """
    Executes the Lava simulation.

    We first compile the templates, initialize Lava and add configured monitors.
    Finally, the compiled Lava code is executed and monitor data is extracted.
    """
    
    # TODO should we use the exec or the import version?
    # NOTE import seems not to work at the moment, exec works
    #import sys
    #sys.path.append(self.project_dir)
    #from process_model import ProcessModel
    #from process import Process
    
    # Compile templates
    processes = compile_templates(self.project_dir, [obj.name for obj in self.lava_objects])

    
    # NOTE adding one step is necessary such that spikes are evaluated
    #      correctly by Lava and match Brian results
    num_steps = int(self.duration/self.defaultclock.dt) + 1

    # First initialize the monitors
    for process in processes.values():
        self.init_lava_monitors(process, num_steps)
    
    # Connect the ports of connected processes
    self.connect_lava_ports(processes)
    root_processes = self.select_root_processes(processes)
    # Log that the run method was called
    self.logger.diagnostic(f'Run Lava simulation for {self.duration} ({num_steps} steps)')
    print(f'Run Lava simulation for {self.duration} ({num_steps} steps)')
    
    # Run the simulation
    for process in root_processes:
        self.logger.debug(f"Running process: {process.name}")
        # Prepare simulation
        run_cfg = Loihi1SimCfg()
        process.run(condition=RunSteps(num_steps=num_steps), run_cfg=run_cfg)

    # After running the simulation, copy the monitored values from all the processes to brian
    for process in processes.values():
        # Set monitor values
        self.set_brian_monitor_values(process)
    
    # Indicate that the network simulation did run
    self.did_run = True


def connect_lava_ports(self, processes):
    """
    Connect the ports of the connected processes.

    Parameters
    ----------
    processes : `Process[]`
        Compiled Lava processes.
    """

    for var in self.lava_ports.values():
        portname = var['portname']
        pathway = var['pathway']

        # First connect the spiking sources to the synapses
        source = processes[var['sender']]
        synapses = processes[pathway.synapses.name]

        # To avoid the DuplicateConnection error from Lava
        syn_spikes_in = synapses.in_ports._members['s_in_'+pathway.prepost]
        # Only reshape the spike port once
        if syn_spikes_in.size == 0:
            # Here we shape the InPort of synapses to accomodate the spikes
            neur_spike_port = processes[source.name].out_ports._members[source.name+'_s_out']
            syn_spikes_in.shape = neur_spike_port.shape
            # Connect the spike ports
            neur_spike_port.connect(syn_spikes_in)
            self.logger.debug(f"Connected {source.name} spiking port to {synapses.name} input")

        # Then connect the synapses to the neurons, here we only require the name
        target = var['receiver']

        # Again, first reshape the ports as needed
        # TODO: not sure this is the intended use of the _members attribute of the Collection class:
        # https://github.com/lava-nc/lava/blob/4283428c3dda02ea6c326d5660a952cafcdd2c03/src/lava/magma/core/process/process.py
        syn_out_port = processes[synapses.name].out_ports._members[portname+'_out']
        neur_in_port = processes[target].in_ports._members[portname+'_in']

        # Reshape the input port of the neuron 
        neur_in_port.shape = syn_out_port.shape
        # Connect the port
        syn_out_port.connect(neur_in_port)

        self.logger.debug(f"Connected {synapses.name} variable {portname} to target neuron: {target}")


def select_root_processes(self, processes):
    """
    This is a helper function to determine which processes to run, in order to avoid running 
    processes connnected to each other twice. Lava automatically runs connected processes together,
    so if we have A-->B-->C we only need to run A.
    This function selects one node from each isolated subnetwork in the lava_ports dictionary through
    a breadth-first search algorithm.

    Parameters
    ----------
    processes : `Process[]`
        A list of the initialized processes in the simulation.

    Returns
    -------
    `list`
        A list of root processes, one from each isolated subnetwork.
    """

    # Build the graph of connected processes
    graph = {}
    for node in self.lava_ports.values():
        sender = node['sender']
        receiver = node['receiver']
        synapses = node['pathway'].synapses.name
        if sender not in graph:
            graph[sender] = set()
        if receiver not in graph:
            graph[receiver] = set()
        graph[sender].add(receiver)
        # Also add the synapses to the sender graph
        graph[sender].add(synapses)
        graph[receiver].add(sender)
    
    # Check for processes which don't require lava ports (isolated components)
    for process in processes:
        if not process in list(graph.keys()):
            graph[process] = set()

    # Log the whole graph to know that everything is working out correctly
    self.logger.diagnostic(graph)

    # Traverse the graph to find connected components
    visited = set()
    components = []
    # BFS algorithm: https://en.wikipedia.org/wiki/Breadth-first_search
    for node in graph:
        if node not in visited:
            component = set()
            queue = [node]
            while queue:
                curr_node = queue.pop(0)
                if curr_node not in visited:
                    visited.add(curr_node)
                    component.add(curr_node)
                    # Add the neighboring nodes to the queue
                    queue.extend(graph[curr_node])
            components.append(component)

    # Select one node from each component
    root_nodes = []
    for component in components:
        root_nodes.append(processes[list(component)[0]])

    return root_nodes


def init_lava_monitors(self, process, num_steps):
    """
    Initializes monitors, defined by the user in Brian.

    Parameters
    ----------
    process : `Process`
        Lava process that is provided with a monitor
    num_steps : `int`
        Number of steps for which the monitor shall be active

    Notes
    -----
    Each probed Lava variable can be assigned to multiple monitors defined in Brian.
    E.g. time can be necessary for a state and a spike monitor, while time is only probed once from Lava.
    """
    # TODO: This might not be the best way to do this, but it ensures that no extra monitors are created
    # and that they are associated with the correct 'sources' only. Probably we could find a more efficient way to do this.
    for m in self.lava_monitors.values():
        # Only look at the right monitors
        if not process.name == m['source']:
            continue

        # If you find a process which should be monitored, look at which variable should be monitored
        lava_var_name = m['lava_var_name']

        # If the given process does not have the required variable, raise an error
        # This should never happen, but just for debugging I'll leave it for now
        if not hasattr(process,lava_var_name):
            raise ValueError(f"Something went wrong: the process {process} does not have the variable {lava_var_name} required from the monitor")
        
        
        # NOTE Lava allows only one monitor per variable, so we only create one if one was not already created
        try:
            # Init Lava monitor and define probe for variable
            monitor = Monitor()
            monitor.probe(getattr(process, lava_var_name), num_steps)
        # If the monitor already exists, lava throws an AssertionError
        except AssertionError:
            # If that's the case, we look for the monitor in the previously defined monitors
            for k in self.lava_monitors.values():
                if not k['lava_monitor'] is None and k['source'] == m['source'] and m['lava_var_name'] == k['lava_var_name']:
                    monitor = k['lava_monitor']
        m['lava_monitor'] = monitor
        m['process_name'] = getattr(process, 'name')


def set_brian_monitor_values(self, process):
    """
    Transform Lava monitors to Brian monitor format.

    Parameters
    ----------
    process : `Process`
        Lava process that is provided with a monitor
    """

    # Iterate over Lava monitors
    for k, m in self.lava_monitors.items():
        # Only do this if the monitor refers to this process
        if not getattr(process,"name") == m['process_name']:
            continue
        # Get monitor variable
        var = m['var']

        # Define empty data variable
        # NOTE Lava monitors do not probe the inital values while Brian does
        #      We need to manually prepend the initial values to the monitored data
        init_data = None  # Contains initial values
        data = None  # Contains all other values

        # In case of a spike monitor, we need a more specific handling to get the time/spike indices
        if m['type'] == self.monitor_types['spike']:

            if var.name == 't':
                # Find monitor that contains 'i' variable
                spike_monitor_name = f'_spike_{var.owner.name}_i'
                spike_monitor = self.lava_monitors[spike_monitor_name]
                # Get raw data from Lava monitor
                raw = get_monitor_values(spike_monitor)
                # Transform data to get actual spike times
                data = (np.nonzero(raw)[0] * self.defaultclock.dt) - self.defaultclock.dt

                # Get initial values from process
                init_raw = getattr(process, f'_{getattr(process,"name")}__spikespace').init
                # Transform data to get actual spike times
                init_data = np.nonzero(init_raw)[0] * self.defaultclock.dt

            if var.name == 'i':
                # Get raw data from Lava monitor
                raw = get_monitor_values(m)
                # Make sure that we always have two dimensions
                if np.ndim(raw) == 1:
                    raw = [raw]
                # Transform data to get indices
                data = np.nonzero(raw)[1]

                # Get initial values from process
                init_raw = getattr(process, f'_{getattr(process,"name")}__spikespace').init
                # Transform initial data to get indices
                init_data = np.nonzero(init_raw)[0]

        # In case of a state monitor, just take the raw data from Lava
        if m['type'] == self.monitor_types['state']:
            # NOTE remove additional simulation steps from monitor to match Brian simulation
            data = get_monitor_values(m)[:-2]

            # Get initial values from process
            init_data = getattr(process, m['lava_var_name']).init

        # Get dimension of data array
        dim = np.ndim(data)

        # Append init values and store data in related array variable
        if dim == 1:
            self.arrays[var] = np.append(init_data, data)

        if dim == 2:
            self.arrays[var] = np.append([init_data], data, axis=0)


def get_monitor_values(monitor):
    """
    Get data from a Lava monitor.

    Parameters
    ----------
    monitor
        A monitor as defined in the device
    """
    # Define variables
    lava_monitor = monitor['lava_monitor']
    lava_var_name = monitor['lava_var_name']
    process_name = monitor['process_name']
    # Return lava monitor values
    # TODO we need to find out how Lava processes are numbered, 'Process_0'
    #      is probably not generalizing for more than one process
    # NOTE: SHOULD BE FIXED NOW!
    data = lava_monitor.get_data()[process_name][lava_var_name]
    return np.squeeze(np.array(data))



from setuptools import setup

setup(
	name='Brian2Lava',
	version='1.0.0a1',  # consider adding '__version__' to 'brian2lava.__init__.py', but requires unified management of versions
	author='Carlo Michaelis, Francesco Negri, Winfried Oed, Jannik Luboeinski, Andrew Lehr, Tristan Stöber',
	author_email='carlo.michaelis@gmail.com',
	packages=['brian2lava', 'brian2lava.device', 'brian2lava.codegen', 'brian2lava.writer'],
	python_requires='>3.8',
	url='https://gitlab.com/tetzlab/brian2lava',
	license='MIT',
	description='An open source Brian2 interface for the neuromorphic computing framework Lava',
	long_description=open('README.md').read(),
	long_description_content_type="text/markdown",
	package_data = {
		"brian2lava/codegen/templates": ["*.py_"],
		"brian2lava/templates": ["*.py.j2"]
	},
    include_package_data=True,
	install_requires=[
		"brian2>=2.5.1",
		"jinja2>=2.7",
		"numpy==1.23.5",
		"pytest",
		"scipy",
		"markupsafe==2.0.1",
		"lava-nc>=0.6.0",
		"matplotlib",
		"regex>=2022.10.31"
	],
)



import itertools

import numpy as np
from regex import sub 

from brian2.parsing.bast import brian_dtype_from_dtype
from brian2.parsing.rendering import NumpyNodeRenderer
from brian2.core.functions import DEFAULT_FUNCTIONS, timestep
from brian2.core.variables import Constant, ArrayVariable
from brian2.utils.stringtools import get_identifiers, word_substitute, indent
from brian2.utils.logger import get_logger
from brian2.core.functions import Function
from brian2.groups.neurongroup import NeuronGroup

from brian2.codegen.generators.base import CodeGenerator


# TODO remove, only for development
from pprint import pprint


__all__ = ['LavaCodeGenerator']


logger = get_logger(__name__)

class VectorisationError(Exception):
    """
    Vectorisation error is used in the code generator
    """

    pass


class LavaCodeGenerator(CodeGenerator):
    """
    Generates Lava code.
    
    Notes
    -----
    Essentially Python but vectorised. It is similar to `NumpyCodeGenerator`, but contains modifications.
    """

    # Define some class variables
    class_name = 'lava'
    _use_ufunc_at_vectorisation = True  # allow this to be off for testing only


    def translate_expression(self, expr):
        """
        TODO
        """

        expr = word_substitute(expr, self.func_name_replacements)
        translated_expr = NumpyNodeRenderer(auto_vectorise=self.auto_vectorise).render_expr(expr, self.variables).strip()
        return translated_expr


    def translate_statement(self, statement):
        """
        Translates abtract code statements to Python code.

        Parameters
        ----------
        statement
            A statement containing the variable, operation, expression and comment.

        Returns
        -------
        `str`
            Compiled code
        """

        # TODO: optimisation, translate arithmetic to a sequence of inplace
        # operations like a=b+c -> add(b, c, a)
        var, op, expr, comment = (statement.var, statement.op, statement.expr, statement.comment)
        if op == ':=':
            op = '='
        # For numpy we replace complex expressions involving a single boolean variable into a
        # where(boolvar, expr_if_true, expr_if_false)
        if (statement.used_boolean_variables is not None and len(statement.used_boolean_variables)==1
                and brian_dtype_from_dtype(statement.dtype)=='float'
                and statement.complexity_std>sum(statement.complexities.values())):
            used_boolvars = statement.used_boolean_variables
            bool_simp = statement.boolean_simplified_expressions
            boolvar = used_boolvars[0]
            for bool_assigns, simp_expr in bool_simp.items():
                _, boolval = bool_assigns[0]
                if boolval:
                    expr_true = simp_expr
                else:
                    expr_false = simp_expr
            code = f'{var} {op} _numpy.where({boolvar}, {expr_true}, {expr_false})'
        else:
            code = f"{var} {op} {self.translate_expression(expr)}"
        if len(comment):
            code += f" # {comment}"
        
        return code


    def ufunc_at_vectorisation(
            self, statement, variables, indices,
            conditional_write_vars, created_vars, used_variables
        ):
        """
        TODO
        """

        if not self._use_ufunc_at_vectorisation:
            raise VectorisationError()
        # Avoids circular import
        from brian2.devices.device import device

        # See https://github.com/brian-team/brian2/pull/531 for explanation
        used = set(get_identifiers(statement.expr))
        used = used.intersection(k for k in list(variables.keys()) if k in indices and indices[k]!='_idx')
        used_variables.update(used)
        if statement.var in used_variables:
            raise VectorisationError()
        expr = NumpyNodeRenderer(auto_vectorise=self.auto_vectorise).render_expr(statement.expr)

        if statement.op == ':=' or indices[statement.var] == '_idx' or not statement.inplace:
            if statement.op == ':=':
                op = '='
            else:
                op = statement.op
            line = f'{statement.var} {op} {expr}'
        elif statement.inplace:
            if statement.op == '+=':
                ufunc_name = '_numpy.add'
            elif statement.op == '*=':
                ufunc_name = '_numpy.multiply'
            elif statement.op == '/=':
                ufunc_name = '_numpy.divide'
            elif statement.op == '-=':
                ufunc_name = '_numpy.subtract'
            else:
                raise VectorisationError()
            array_name = device.get_array_name(variables[statement.var])
            idx = indices[statement.var]
            line = f'{ufunc_name}.at({array_name}, {idx}, {expr})'
            line = self.conditional_write(line, statement, variables,
                conditional_write_vars=conditional_write_vars, created_vars=created_vars)
        else:
            raise VectorisationError()

        if len(statement.comment):
            line += f" # {statement.comment}"

        return line


    def vectorise_code(self, statements, variables, variable_indices, index='_idx'):
        """
        Translates abtract code statements to vectorized Python code.

        Parameters
        ----------
        statements
            Statements containing the corresponding variable, operation, expression and comment.
        variables
            All variables to compile.
        variable_indices
            If a statement only applies to a subset of the variable (e.g. a subset of a list).
            The subset of the variable is chosen by indices, given by `variable_indices`.
        index
            Name of the index variable. Default: `_idx`

        Returns
        -------
        `str`
            Lines of compiled code
        """

        created_vars = {stmt.var for stmt in statements if stmt.op == ':='}
        try:
            lines = []
            used_variables = set()
            for statement in statements:
                lines.append(f'#  Abstract code:  {statement.var} {statement.op} {statement.expr}')
                # We treat every statement individually with its own read and write code
                # to be on the safe side
                read, write, indices, conditional_write_vars = self.arrays_helper([statement])
                # We make sure that we only add code to `lines` after it went
                # through completely
                ufunc_lines = []
                # No need to load a variable if it is only in read because of
                # the in-place operation
                if (statement.inplace and
                            variable_indices[statement.var] != '_idx' and
                            statement.var not in get_identifiers(statement.expr)):
                    read = read - {statement.var}
                ufunc_lines.extend(
                    self.read_arrays(read, write, indices, variables, variable_indices)
                )
                ufunc_lines.append(
                    self.ufunc_at_vectorisation(
                        statement, variables, variable_indices, conditional_write_vars,
                        created_vars, used_variables,
                    )
                )
                # Do not write back such values, the ufuncs have modified the
                # underlying array already
                if statement.inplace and variable_indices[statement.var] != '_idx':
                    write = write - {statement.var}
                ufunc_lines.extend(
                    self.write_arrays([statement], read, write, variables, variable_indices)
                )
                lines.extend(ufunc_lines)
        except VectorisationError:
            if self._use_ufunc_at_vectorisation:
                logger.info("Failed to vectorise code, falling back on Python loop: note that "
                            "this will be very slow! Switch to another code generation target for "
                            "best performance (e.g. cython). First line is: "+str(statements[0]),
                            once=True)
            lines = []
            lines.extend(['_full_idx = _idx',
                          'for _idx in _full_idx:',
                          '    _vectorisation_idx = _idx'
                          ])
            read, write, indices, conditional_write_vars = self.arrays_helper(statements)
            lines.extend(indent(code) for code in
                         self.read_arrays(read, write, indices,
                                          variables, variable_indices))
            for statement in statements:
                line = self.translate_statement(statement)
                if statement.var in conditional_write_vars:
                    lines.append(indent(f'if {conditional_write_vars[statement.var]}:'))
                    lines.append(indent(line, 2))
                else:
                    lines.append(indent(line))
            lines.extend(indent(code) for code in
                         self.write_arrays(statements, read, write,
                                           variables, variable_indices))
        
        return lines


    @staticmethod
    def get_array_name(var, access_data=True, template_name=None):
        """
        Returns a variable name used in the template
        Specifically differentiates between the usage of init and lava variables
        It uses the devices `get_array_name` method

        Notes
        -----
        It is important that this function has the name 'get_array_name', since it is also called by
        internal Brian functions and overwrites the default 'get_array_name' in CodeGenerator
        """

        # We have to do the import here to avoid circular import dependencies
        from brian2.devices.device import get_device
        # Get device
        device = get_device()

        # Template name is not specified
        if template_name is None:
            # '_group_idx' is a special case, see e.g. 'group_variable_set.py_'
            # We assume that '_group_idx' is only used in *process* related templates
            if '_group_idx' in var.name:
                return device.get_array_name(var, access_data, prefix='self.init')
            else:
                return device.get_array_name(var, access_data)
        # If variable is related to a template that is used in *process*
        elif template_name in device.init_template_functions:
            return device.get_array_name(var, access_data, prefix='self.init')
        # if variable is related to a template that is used in *process model*
        else:
            return device.get_array_name(var, access_data)


    def read_arrays(self, read, write, indices, variables, variable_indices):
        """
        Create statements that read from array.

        Parameters
        ----------
        read
            Arrays names to read from.
        write
            Arrays names to write to.
        indices
            Indices of arrays to read from.
        variables
            Actual variables that can be selected by variable names (e.g. contained in `read`).
        variable_indices
            Indices to choose ranges from the variables.

        Returns
        -------
        `str`
            Lines of compiled code
        """

        # index and read arrays (index arrays first)
        lines = []
        for varname in itertools.chain(indices, read):
            var = variables[varname]
            index = variable_indices[varname]
            if self.template_name == 'activation_processing' and 'synapses' in var.owner.name:
                line = f"{varname} = {varname}_received"
            else:
                line = f"{varname} = {self.get_array_name(var, template_name = self.template_name)}"
            
            if index not in self.iterate_all:
                line += f"[{index}]"
            elif varname in write:
                # avoid potential issues with aliased variables, see github #259
                line += '.copy()'
            lines.append(line)

        return lines


    def write_arrays(self, statements, read, write, variables, variable_indices):
        """
        Create statements that write to array.

        Parameters
        ----------
        statements
            Statements containing the corresponding variable, operation, expression and comment.
        read
            Arrays names to read from.
        write
            Arrays names to write to.
        variables
            Actual variables that can be selected by variable names (e.g. contained in `write`).
        variable_indices
            Indices to choose ranges from the variables.

        Returns
        -------
        `str`
            Lines of compiled code
        """

        lines = []
        for varname in write:
            var = variables[varname]
            index_var = variable_indices[varname]
            # check if all operations were inplace and we're operating on the
            # whole vector, if so we don't need to write the array back
            if index_var not in self.iterate_all or varname in read:
                all_inplace = False
            else:
                all_inplace = True
                for stmt in statements:
                    if stmt.var == varname and not stmt.inplace:
                        all_inplace = False
                        break
            if not all_inplace:
                line = self.get_array_name(var, template_name = self.template_name)

                if index_var in self.iterate_all:
                    line = f"{line}[:]"
                else:
                    line = f"{line}[{index_var}]"
                line = f"{line} = {varname}"
                lines.append(line)

        return lines


    def conditional_write(self, line, stmt, variables, conditional_write_vars, created_vars):
        """
        TODO
        """

        if stmt.var in conditional_write_vars:
            subs = {}
            index = conditional_write_vars[stmt.var]
            # we replace all var with var[index], but actually we use this repl_string first because
            # we don't want to end up with lines like x[not_refractory[not_refractory]] when
            # multiple substitution passes are invoked
            # FIXME let's have another look at this at some point
            repl_string = '#$(@#&$@$*U#@)$@(#'  # this string shouldn't occur anywhere I hope! :)
            for varname, var in list(variables.items()):
                if isinstance(var, ArrayVariable) and not var.scalar:
                    subs[varname] = f"{varname}[{repl_string}]"
            # all newly created vars are arrays and will need indexing
            for varname in created_vars:
                subs[varname] = f"{varname}[{repl_string}]"
            # Also index _vectorisation_idx so that e.g. rand() works correctly
            subs['_vectorisation_idx'] = f"_vectorisation_idx[{repl_string}]"

            line = word_substitute(line, subs)
            line = line.replace(repl_string, index)

        return line


    def translate_statement_sequence(self, scalar_statements, vector_statements):
        """
        Takes lines of code and defines Jinja variables, stored in Python dictionaries.

        Parameters
        ----------
        scalar_statements
            Lines of code containing *scalar* statements.
        vector_statements
            Lines of code containing *vector* statements.

        Returns
        -------
        scalar_code : `dict`
            Jinja variables containing compiled *scalar* code.
        vector_code : `dict`
            Jinja variables containing compiled *vector* code.
        kwds : `dict`
            Jinja variables containing additional custom variables (so called 'keywords').
        """

        scalar_code = {}
        vector_code = {}
        for name, block in scalar_statements.items():
            scalar_code[name] = self.translate_one_statement_sequence(
                block, scalar=True
            )
        for name, block in vector_statements.items():
            vector_code[name] = self.translate_one_statement_sequence(
                block, scalar=False
            )

        kwds = self.determine_keywords()

        return scalar_code, vector_code, kwds


    def translate_one_statement_sequence(self, statements, scalar=False):
        """
        Translate an abstract code statement sequence to Python code.

        Parameters
        ----------
        statements
            Statements containing the corresponding variable, operation, expression and comment.
        scalar: `bool`
            Indicates if the statements shall be scalar or vectorized.

        Returns
        -------
        lines : `str[]`
            List of compiled code lines.
        """

        variables = self.variables
        variable_indices = self.variable_indices
        read, write, indices, conditional_write_vars = self.arrays_helper(statements)
        lines = []
        all_unique = not self.has_repeated_indices(statements)
        self.check_for_learning_statements(statements)
        if scalar or all_unique:
            # Simple translation
            lines.extend(self.read_arrays(read, write, indices, variables, variable_indices))
            created_vars = {stmt.var for stmt in statements if stmt.op == ':='}
            for stmt in statements:
                line = self.translate_statement(stmt)
                line = self.conditional_write(line, stmt, variables, conditional_write_vars, created_vars)
                lines.append(line)
            lines.extend(self.write_arrays(statements, read, write, variables, variable_indices))
        else:
            # More complex translation to deal with repeated indices
            lines.extend(self.vectorise_code(statements, variables, variable_indices))

        return lines


    def check_for_learning_statements(self, statements):
        """
        A method to look for learning statements and implement learning rules in the way of lava

        Parameters
        ----------
        statements
            Statements containing the corresponding variable, operation, expression and comment.
        """

        # For release this is just going to throw an error, as learning is not implemented yet
        for stmt in statements:
            if self.template_name == 'activation_processing' and not isinstance(self.variables[stmt.var].owner,NeuronGroup):
                msg = f""" In the current release learning is not supported yet: the 'on_pre' and 'on_post' conditions
                can only be used to update variables belonging to NeuronGroups. (e.g. on_pre = 'v+=w' or similar).
                The following expression is not allowed: {stmt}
                """
                raise NotImplementedError(msg)


    def determine_keywords(self):
        """
        Adds custom variables to the Jinja templates
        """

        from brian2 import get_device
        device = get_device()

        # Add constants to template keywords
        # TODO Francesco: I'm wondering if this is the correct way to add the 'USES_VARIABLES' from templates
        # check: https://github.com/brian-team/brian2/blob/3cf65d4c5c8fa1512cf0a6deb515a8cdc81ae9e5/brian2/codegen/templates.py (186-217)
        # It seems that the templates themselves should take care of this. Further investigation is needed.
        constants = []
        for var in self.variables.values():
            if isinstance(var, Constant):
                constants.append(f'{var.name} = {var.get_value()}')
            elif isinstance(var, ArrayVariable) and var.constant:
                constants.append(f'{var.name} = {device.get_array_name(var)}[0]')
        read_syn_vars = []
        syn_output_vars = []
        syn_output_ports = []

        # Read the required synaptic variables which are to be sent to neurons
        # and send them out through the correct port.
        if self.template_name == 'synapses':            
            ports = device.lava_ports
            for port in ports.values():
                if not port['pathway'].synapses == self.owner or not port['varname'] in list(self.variables.keys()) or port['pathway'].prepost not in self.name:
                    continue
                varname = port['varname']
                lava_name = device.get_array_name(self.owner.variables[varname])
                # Not sure about this
                _idx = self.variable_indices[varname]
                read_syn_vars.append(f'{varname} = {lava_name}[{_idx}]')
                syn_output_vars.append(varname)
                syn_output_ports.append(port['portname'])

        # Receive the input from the synapses
        read_port_input = []
        neur_input_vars = []
        if self.template_name == 'activation_processing':
            from brian2 import get_device
            device = get_device()
            ports = device.lava_ports
            for port in ports.values():
                if not port['receiver'] == self.owner.name or not port['varname'] in list(self.variables.keys()):
                    continue
                varname = port['varname'] + '_received'
                portname = port['portname']
                # Not sure about this
                read_port_input.append(f'{varname} = self.{portname}_in.recv()')
                neur_input_vars.append(varname)

        # Determine if scipy is available
        try:
            import scipy
            scipy_available = True
        except ImportError:
            scipy_available = False

        # Collect template keywords and return
        return {
            '_scipy_available': scipy_available,
            'constants': constants,
            'read_port_input': read_port_input,
            'neur_input_vars': neur_input_vars,
            'syn_output_vars': syn_output_vars,
            'syn_output_ports': syn_output_ports,
            'read_syn_vars': read_syn_vars,
            'zip': zip
        }


# Functions that exist under the same name in numpy
for func_name, func in [
    ('sin', np.sin), ('cos', np.cos), ('tan', np.tan), ('sinh', np.sinh),
    ('cosh', np.cosh), ('tanh', np.tanh), ('exp', np.exp), ('log', np.log),
    ('log10', np.log10), ('sqrt', np.sqrt), ('arcsin', np.arcsin),
    ('arccos', np.arccos), ('arctan', np.arctan), ('abs', np.abs), ('sign', np.sign)
]:
    DEFAULT_FUNCTIONS[func_name].implementations.add_implementation(LavaCodeGenerator, code=func)

# Functions that are implemented in a somewhat special way
def randn_func(vectorisation_idx):
    try:
        N = len(vectorisation_idx)
        return np.random.randn(N)
    except TypeError:
        # scalar value
        return np.random.randn()


def rand_func(vectorisation_idx):
    try:
        N = len(vectorisation_idx)
        return np.random.rand(N)
    except TypeError:
        # scalar value
        return np.random.rand()


def poisson_func(lam, vectorisation_idx):
    try:
        N = len(vectorisation_idx)
        return np.random.poisson(lam, size=N)
    except TypeError:
        # scalar value
        return np.random.poisson(lam)


DEFAULT_FUNCTIONS['randn'].implementations.add_implementation(LavaCodeGenerator, code=randn_func)
DEFAULT_FUNCTIONS['rand'].implementations.add_implementation(LavaCodeGenerator, code=rand_func)
DEFAULT_FUNCTIONS['poisson'].implementations.add_implementation(LavaCodeGenerator, code=poisson_func)
clip_func = lambda array, a_min, a_max: np.clip(array, a_min, a_max)
DEFAULT_FUNCTIONS['clip'].implementations.add_implementation(LavaCodeGenerator, code=clip_func)
int_func = lambda value: np.int32(value)
DEFAULT_FUNCTIONS['int'].implementations.add_implementation(LavaCodeGenerator, code=int_func)
ceil_func = lambda value: np.int32(np.ceil(value))
DEFAULT_FUNCTIONS['ceil'].implementations.add_implementation(LavaCodeGenerator, code=ceil_func)
floor_func = lambda value: np.int32(np.floor(value))
DEFAULT_FUNCTIONS['floor'].implementations.add_implementation(LavaCodeGenerator, code=floor_func)

# We need to explicitly add an implementation for the timestep function,
# otherwise Brian would *add* units during simulation, thinking that the
# timestep function would not work correctly otherwise. This would slow the
# function down significantly.
DEFAULT_FUNCTIONS['timestep'].implementations.add_implementation(LavaCodeGenerator, code=timestep)



import numpy as np

from brian2.core.variables import Variables


def variableview_set_with_index_array(self, variableview, item, value, check_units):
    """
    Is called when a variable is set with a numerical (or slice) index and numerical target value (e.g. G.v[5:] = -70*mV)

    Parameters
    ----------
    variableview
        Corresponds to `self` within the orginal variableview
    item : `ndarray`
        The indices for the variable (in the context of this `group`).
    value : value
        A Brian value and unit
    check_units : bool, optional
        Whether to check the units of the expression.
    """

    # Update array variable
    self.fill_with_array(variableview.variable, value)


def variableview_set_with_expression_conditional(self, variableview, cond, code, run_namespace, check_units=True):
    """
    Is called when a variable is set with a string index and a string expression value (e.g. G.v['i>5'] = (-70+i)*mV)

    Parameters
    ----------
    variableview 
        Corresponds to `self` within the orginal variableview
    cond : str
        The string condition for which the variables should be set.
    code : str
        The code that should be executed to set the variable values.
        Can contain references to indices, such as `i` or `j`
    run_namespace : dict-like, optional
        An additional namespace that is used for variable lookup (if not
        defined, the implicit namespace of local variables is used).
    check_units : bool, optional
        Whether to check the units of the expression.
    """

    # Try to cast code to value
    value = None
    try:
        value = int(code)
    except:
        raise NotImplementedError('Defining variables as expressions is currently not supported. For more information: https://gitlab.com/tetzlab/brian2lava/-/issues/19')
    # Update array variable
    self.fill_with_array(variableview.variable, value)


def variableview_set_with_expression(self, variableview, item, code, run_namespace, check_units=True):
    """
    Sets a variable using a string expression. Is called by `VariableView.set_item` for statements such as
    ``S.var[:, :] = 'exp(-abs(i-j)/space_constant)*nS'``.

    Parameters
    ----------
    variableview
        Corresponds to `self` within the orginal variableview
    item : `ndarray`
        The indices for the variable (in the context of this `group`).
    code : str
        The code that should be executed to set the variable values.
        Can contain references to indices, such as `i` or `j`
    run_namespace : dict-like, optional
        An additional namespace that is used for variable lookup (if not
        defined, the implicit namespace of local variables is used).
    check_units : bool, optional
        Whether to check the units of the expression.
    """

    # Some fairly complicated code to raise a warning in ambiguous
    # situations, when indexing with a group. For example, in:
    #   group.v[subgroup] =  'i'
    # the index 'i' is the index of 'group' ("absolute index") and not of
    # subgroup ("relative index")
    if hasattr(item, "variables") or (
        isinstance(item, tuple)
        and any(hasattr(one_item, "variables") for one_item in item)
    ):
        # Determine the variables that are used in the expression
        from brian2.codegen.translation import get_identifiers_recursively

        identifiers = get_identifiers_recursively([code], variableview.group.variables)
        variables = variableview.group.resolve_all(
            identifiers, run_namespace, user_identifiers=set()
        )
        if not isinstance(item, tuple):
            index_groups = [item]
        else:
            index_groups = item

        for varname, var in variables.items():
            for index_group in index_groups:
                if not hasattr(index_group, "variables"):
                    continue
                if (
                    varname in index_group.variables
                    or var.name in index_group.variables
                ):
                    indexed_var = index_group.variables.get(
                        varname, index_group.variables.get(var.name)
                    )
                    if indexed_var is not var:
                        self.logger.warn(
                            "The string expression used for setting "
                            f"'{variableview.name}' refers to '{varname}' which "
                            "might be ambiguous. It will be "
                            "interpreted as referring to "
                            f"'{varname}' in '{variableview.group.name}', not as "
                            "a variable of a group used for "
                            "indexing.",
                            "ambiguous_string_expression",
                        )
                        break  # no need to warn more than once for a variable
    
    indices = np.atleast_1d(variableview.indexing(item))
    abstract_code = f"{variableview.name} = {code}"
    variables = Variables(variableview.group)

    # Get number of set variable indices to be able to set unique names
    counter = len(self.set_variable_index_names)+1
    # Get group idx name for current index iteration
    idx_name = f'_group_idx_{variableview.name}_{counter}'

    # Add group idx variable
    # This function is part of the Variables class, but also calles get_array_name at some point
    variables.add_array(idx_name, size=len(indices), dtype=np.int32, values=indices)

    # Add the new group idx name to an array (needed for the counter)
    self.set_variable_index_names.append(variables._variables[idx_name])

    # Store method name and lava variable name in init variables
    # and set the current method name as last-used name (needed for the code object to identify the method name)
    method_name = f'{variableview.group_name}_group_variable_set_{counter}_{variableview.name}'
    lava_variable_name = self.get_array_name(variableview.variable, prefix='')
    self.init_variables[method_name] = lava_variable_name
    self.last_set_variable_method_name = method_name

    # Needs to be added to namespace, since group still requires the original name '_group_idx'
    # The specific reason for that is unknown ;)
    run_namespace['_group_idx'] = variables._variables[idx_name]

    # TODO: Have an additional argument to avoid going through the index
    # array for situations where iterate_all could be used
    from brian2.codegen.codeobject import create_runner_codeobj

    codeobj = create_runner_codeobj(
        variableview.group,
        abstract_code,
        "group_variable_set",
        additional_variables=variables,
        check_units=check_units,
        run_namespace=run_namespace,
        codeobj_class=self.code_object_class(
            fallback_pref="codegen.string_expression_target"
        ),
    )



import os

# Import Brian2 modules
from brian2.utils.logger import get_logger


class PyWriter(object):
    '''
    Writes given contents to Python files.
    Inspired from CPPWriter from the cpp_standalone device.
    '''

    def __init__(self, project_directory):
        '''
        Initializes the project directory
        '''

        # Store project directory
        self.project_directory = project_directory

        # Make logger available to all methods in this class
        self.logger = get_logger('brian2.devices.lava')

    def write(self, filename, contents):
        '''
        Writes a given content to a file with given filename
        '''

        # Log contents and filename
        self.logger.diagnostic(f'Writing file {filename}:\n{contents}')

        # If filename does not end with .py, raise exception
        if not filename.lower().endswith('.py'):
            ValueError("Extension of your Python file must be .py")

        # Assemble path
        path = os.path.join(self.project_directory, filename)

        # If path already exist and content of the file matches the given content, return
        if os.path.exists(path):
            with open(path, 'r') as f:
                if f.read() == contents:
                    return

        # Write contents to file
        with open(path, 'w') as f:
            f.write(contents)
